/*
** Module   : SymbolResolver.java
** Abstract : manages the symbol dictionaries and lookup process
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20041104   @18650 First version, with support for resolving:
**                           - optionally abbreviated keywords
**                           - variables with multiple scoping levels
**                             including a special global scope
** 002 GES 20041115   @18651 Added support for:
**                           - functions
**                           - procedures
**                           - labels
** 003 GES 20041208   @18937 Added new methods for lookup of variables
**                           and functions.  These new methods are now
**                           using the old "default" names and the old
**                           methods that handle exact matching are under
**                           new names lookup*Exact().  The new methods 
**                           implement an algorithm that checks to see
**                           if a match is made to a language keyword,
**                           if such a match exists, a 2nd level of
**                           processing occurs. Otherwise the processing
**                           is the same as before.  This new algorithm
**                           is necessary to allow built-in functions and
**                           variables to be properly resolved, since
**                           these symbols are keywords which must support
**                           the reserved/unreserved attribute and the
**                           matching must be done considering
**                           abbreviations.
** 004 GES 20041217   @19041 Added a first pass implementation for the
**                           lookup of database objects (databases,
**                           tables and fields) in both unqualified
**                           and partially/fully qualified forms.  In
**                           addition, table and field abbreviations are
**                           fully supported.  Warning: table support is
**                           hard coded at this time.                               
** 005 NVS 20041225   @19109 Now the label dictionary is scoped, with one
**                           label per scope (Progress allows only one
**                           label per block). Modified addLabel creates
**                           new scope implicitly. Added deleteLabel()
**                           method which deletes the current scope.
** 006 ECF 20050103   @19188 Reimplemented lookupDatabase, lookupTable,
**                           and lookupField methods to handle exceptions
**                           thrown by SchemaDictionary. In each case, if
**                           an exception is caught, error information is
**                           written to stdout and -1 is returned as the
**                           token type.
** 007 GES 20050103   @19192 Added isTable and isField methods which 
**                           don't throw exceptions. This allows the
**                           parser to test things that may be ambiguous
**                           without being hit over the head and left
**                           bleeding in the gutter.
** 008 GES 20050107   @19270 Widgets are handled just like variables 
**                           (they share the same namespace) in Progress.
**                           However due to the fact that some symbols
**                           do "double duty" as both variables and as
**                           widgets, a separate widget namespace is
**                           kept so that the generic one name to one
**                           token type mapping can be kept for the
**                           majority of the standard name lookups. A
**                           new widgetDict member is added and two
**                           methods (addWidget and lookupWidget) are
**                           provided to provide access.  This new
**                           namespace does participate in the scoping
**                           just like variables.
** 009 GES 20050118   @19336 Added support for modifying the schema
**                           namespace (adding tables, fields and lists
**                           of fields).
** 010 GES 20050119   @19346 Added support for record scoping (driving
**                           the schema's scopes and which tables are
**                           promoted to the current scope).
** 011 GES 20050127   @19449 Added new namespaces for frames, menus (+
**                           sub-menus) and menu-items.
** 012 GES 20050128   @19460 Added new namespace for streams.
** 013 GES 20050201   @19513 Added additional parser token type to
**                           schema dictionary data type conversions
**                           (for variable types like VAR_INT). Also
**                           added dump support for the schema dictionary.
** 014 GES 20050204   @19570 Added loadSchemaDatabase() method to load
**                           an arbitrary non-default database into the
**                           schema namespace.
** 015 GES 20050207   @19624 Removed deprecated method call to schema
**                           namespace.
** 016 GES 20050209   @19737 Modified addTable() method to allow user
**                           control over where a table is created,
**                           either in the global scope or in an internal
**                           procedure or function scope (buffers only).
** 017 GES 20050210   @19749 Added the ability to add a variable using
**                           data contained in a wrapper class. This may
**                           be extended later to store more data in the
**                           dictionary than is possible today. In
**                           addition a new method of adding lists of
**                           variables, widgets and frames to the
**                           respective dictionaries was provided. Also
**                           added a method for adding aliases to the 
**                           schema namespace.
** 018 GES 20050218   @19860 Minor improvements to the loading of
**                           variables, to allow fields and a special
**                           token type of DYNAMIC to be loaded by the
**                           same mechanism.
** 019 GES 20050307   @20217 Added methods to lookup the fully qualified
**                           name of a field or table.
** 020 GES 20050331   @20561 Major additions to store options at the
**                           time that variables and temp-tables are
**                           defined. These options are saved in the
**                           variable dictionary in the case of vars and
**                           are pushed out to the schema dictionary in
**                           the case of temp-tables. All variable and
**                           temp-table definitions that are LIKE another
**                           variable, field or table now honor and
**                           bidirectionally pass the options as needed.
**                           Variable and field references (the lvalue
**                           rule) store annotations for non-default
**                           options. WARNING: INITIAL, LABEL and
**                           VIEW-AS are 3 common options that are not
**                           processed or are only partially processed
**                           at this time.
** 021 ECF 20050331   @20613 Implemented all schema dictionary hooks added
**                           for 020 (@20561).
** 022 GES 20050401   @20596 Added support for a temporary index to allow
**                           the linking of a variable reference to its
**                           original definition at parse time. This is
**                           needed because project-unique IDs can't be
**                           easily generated at parse time, but are only
**                           added after the entire tree is complete.
** 023 GES 20050406   @20651 Massive rewrite to use ASTs to store variable
**                           options and to exchange ASTs with the schema
**                           dictionary and fully handle field and
**                           variable options.  In addition it is now
**                           possible to instantiate this class in a
**                           manner that bypasses the schema dictionary.
**                           This allows the ProgressLexer to be used
**                           without dependencies upon an already
**                           functional schema.
** 024 ECF 20050413   @20701 Implemented all schema dictionary hooks which
**                           needed to be modified as a result of 023
**                           (@20651). TODO: still need to handle VALIDATE
**                           keyword.
** 025 GES 20050414   @20723 Converted function processing into a wrapped
**                           approach that allows storing other attributes
**                           and providing annotation helpers.  This
**                           required some rework of the variable/field
**                           processing.
** 026 GES 20050509   @21132 Modified annotateVariableOptions() to be
**                           safe when called by the class
**                           ProgressExpressionEvaluator which short-
**                           circuits the normal variable dictionary
**                           processing and thus can generate nulls
**                           where they are normally not possible.
** 027 GES 20050822   @22187 Support schema dictionary record type
**                           lookups for fields and buffers.  This
**                           reports the token type of the backing
**                           construct (table, temp-table or work-table).
** 028 GES 20050830   @22431 Added attribute/method namespace support.
** 029 GES 20050922   @22778 Support schema dictionary buffer name
**                           lookups for fields and records.  This is
**                           sometimes (often) different from the fully
**                           qualified schema name and is needed to know
**                           which buffer this reference is uniquely
**                           referencing.
** 030 GES 20051004   @22931 Added database name lookup for tables (this
**                           is sometimes different from the fully
**                           qualified schema name).
** 031 GES 20051013   @23058 Moved field annotation here from parser and
**                           added helper to create a field node based
**                           on an unambiguous field name.
** 032 GES 20051109   @23282 Added more annotations and tempidx support
**                           for non-var widgets (buttons/browse) by
**                           using the wrapper approach.
** 033 ECF 20060425   @25740 Enhanced support for define temp-table. A
**                           field can be like a specific element of an
**                           extent field.
** 034 ECF 20060427   @25823 Enhanced support for defining a variable like
**                           a database field. Enabled specification of a
**                           specific element of an extent field as the
**                           model field.
** 035 GES 20070329   @32655 Extended the concept in H034 to all forms of
**                           variable and field definition using another
**                           variable.
** 036 GES 20070404   @32793 Pass through a marker when adding a scope to
**                           the schema dictionary. This allows the new
**                           scope to be associated with the fact of
**                           whether the scope belongs to an internal
**                           procedure, user-defined function or trigger.
** 037 GES 20070418   @33112 Added support for COM-HANDLE types.
** 038 GES 20070528   @33788 Added access to function definition object.
** 039 GES 20070629   @34346 Changed APIs to pass through enough ASTs
**                           to allow the schema dictionary to set the
**                           line and column info for temp-table ASTs
**                           created during parsing.
** 040 GES 20070907   @35010 Support for 10.1B data types.
** 041 GES 20070911   @35138 Support for 10.1B class and package
**                           namespaces.
** 042 GES 20070923   @35184 Added removeKeyword().
** 043 GES 20070927   @35257 Modified promotion behavior to match new
**                           interfaces in the schema dictionary. Added
**                           table promotion when qualified field
**                           references are encountered.
** 044 GES 20071012   @35529 Added separate namespaces for dataset,
**                           data-source and data-relation resources.
** 045 GES 20071107   @35900 Made most schema related methods safe to
**                           use with a null schema dictionary. This
**                           allows the parser to operate on a wide range
**                           of expressions without any knowledge or
**                           dependency upon the schema dictionary. This
**                           is important in cases of preprocessor
**                           expression evaluation, where text may need
**                           to be parsed as a symbol instead of as a
**                           field name (but if the schema dictionary
**                           was loaded, the preprocessor symbol would
**                           resolve as a field which is an invalid
**                           result).
** 046 GES 20071129   @36187 Removed support for DYNAMIC hints (no longer
**                           needed).
** 047 GES 20071207   @36286 Signature changes to pass more data to the
**                           schema dictionary. Also added a global flag
**                           to enable/disable table promotion.
** 048 ECF 20080704   @39145 Minor memory optimization. Use Boolean
**                           constants instead of instantiating new
**                           Boolean objects.
** 049 GES 20090429   @42065 Match package and class name changes.
** 050 GES 20110712          Made failure to find a class abend more gracefully
**                           (and provide more useful data).
** 051 GES 20110811          Added simple list of handles that could be possible
**                           active-x targets. This should probably be a real
**                           scoped dictionary, but for a start it should handle
**                           virtually all cases.
** 052 GES 20110822          Added support for loading built-in 4GL OO classes/
**                           interfaces and .NET classes/interfaces. Added
**                           static support. Added support for define event
**                           when loading a class via AST. Events are treated
**                           like properties and vars (they share the same
**                           namespace). Reworked class loading for prescanning
**                           and to improve performance.
** 053 GES 20111005          Added safety code to avoid class loading issues
**                           when the OO skeleton class directories don't exist.
** 054 CA  20130225          Annotate the fields with the db name.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.schema.*;
import com.goldencode.p2j.util.*;

/**
 * Contains a dictionary for each type of language namespace and provides
 * methods to maintain and lookup in the respective namespaces.
 * <p>
 * Provides support for the following namespaces:
 * <ul>
 *    <li> language keywords (optionally abbreviated)
 *    <li> global and local variables with multiple arbitrary levels of
 *         scoping
 *    <li> functions (both built-in and user-defined)
 *    <li> procedures
 *    <li> classes
 *    <li> package names
 *    <li> labels
 *    <li> schema names
 *    <ul>
 *       <li> databases
 *       <li> tables - qualified or not, optionally abbreviated
 *       <li> fields - qualified or not, optionally abbreviated
 *    </ul>
 *    <li> widgets (this namespace is artificially separated from the
 *         variable namespace though normally this is not the case in
 *         Progress)
 *    <li> frames 
 *    <li> menus and sub-menus
 *    <li> menu-items
 *    <li> streams
 *    <li> ProDataSet support
 *    <ul> 
 *       <li> datasets
 *       <li> data-sources
 *       <li> data-relations
 *    </ul> 
 *    <li> object-oriented support
 *    <ul> 
 *       <li> packages
 *       <li> class definitions
 *    </ul> 
 * </ul>
 * <p>
 * All namespaces (except language keywords) provide a mapping between a 
 * symbol and an integer token type as defined by the Lexer and Parser.  It
 * is this token type that is returned from most lookup operations.
 * <p>
 * This class is aware of the parser token types and where necessary it
 * converts other values into the proper types for use in the parser.
 * At this time this feature is only needed for schema name lookups as
 * all other types are already normalized properly in the respective 
 * dictionary.  See {@link #lookupField}.
 * <p>
 * Only the variable, widget and label namespaces utilize multiple scoping
 * levels at this time.  All other namespaces are flat.
 */
public class SymbolResolver
implements ProgressParserTokenTypes
{
   /** Possible access modes for object data members. */
   private static final Set<Integer> ACCESS_MODES = new HashSet<Integer>();
   
   /** Class or interface definition file extension. */
   public static final String CLASS_EXT = ".cls";
   
   /** Class or interface definition file search specification. */
   public static final String CLASS_SPLAT = "*" + CLASS_EXT;
   
   /** Class or interface definition file extension. */
   public static final String CLASS_MAP_FILE = "./cfg/class_map.xml";
   
   /** Default parent of all classes if no explicit parent is specified. */
   public static final String ROOT_OBJ_NAME = "Progress.Lang.Object";
   
   /** Default parent of all classes if no explicit parent is specified. */
   public static final String ROOT_CLASS_NAME = "Progress.Lang.Class";
   
   /** Relative location for the non-project class/interface 4GL code. */
   public static final String SKELETON_PATH = "./p2j/skeleton/";
   
   /** Relative location for the 4GL object-oriented base definitions. */
   public static final String OO4GL_PATH = SKELETON_PATH + "oo4gl/";
   
   /** Relative location for the .NET base definitions. */
   public static final String DOTNET_PATH = SKELETON_PATH + "dotnet/";
   
   /** Relative location for project-specific .NET definitions. */
   public static final String ASSEMBLY_PATH = "./assemblies/";
   
   /** Stores class and interface definitions. */
   private static ScopedSymbolDictionary classDict = null;
   
   /** Classes in the propath, mapped qualified classname to filename. */
   private static Map<String, String> propathCls = null;
   
   /** Classes in the OO4GL_PATH, mapped qualified classname to filename. */
   private static Map<String, String> oo4glCls = null;
   
   /** Classes in the DOTNET_PATH, mapped qualified classname to filename. */
   private static Map<String, String> dotnetCls = null;
   
   /** Classes in the ASSEMBLY_PATH, mapped qualified classname to filename. */
   private static Map<String, String> assemblyCls = null;
   
   /** The most recently used propath for init of the propath class map. */
   private static String[] lastPropath = null;
   
   /** Stores keywords and handles abbreviation support. */
   private KeywordDictionary kwDict = null;
   
   /** Stores variables, both local (scoped) and global. */
   private ScopedSymbolDictionary varDict = null;
   
   /** Stores function names and return types. */
   private ScopedSymbolDictionary funcDict = null;
   
   /** Stores procedure names. */
   private ScopedSymbolDictionary procDict = null;
   
   /** Stores explicit package names. */
   private Map<String, UsingSpec> explPkgDict = null;
   
   /** Stores package search specifications. */
   private Set<UsingSpec> searchPkgDict = null;
   
   /** Stores label names. */
   private ScopedSymbolDictionary labelDict = null;   
   
   /** Stores database object names (databases, tables and fields). */
   private SchemaDictionary schemaDict = null;
   
   /** Stores widget names. */
   private ScopedSymbolDictionary widgetDict = null;
   
   /** Stores frame names. */
   private ScopedSymbolDictionary frameDict = null;
   
   /** Stores menu and sub-menu names. */
   private ScopedSymbolDictionary menuDict = null;
   
   /** Stores menu-item names. */
   private ScopedSymbolDictionary menuItemDict = null;
   
   /** Stores stream names. */
   private ScopedSymbolDictionary streamDict = null;
   
   /** Stores dataset names. */
   private ScopedSymbolDictionary dsetDict = null;
   
   /** Stores data-source names. */
   private ScopedSymbolDictionary dsrcDict = null;
   
   /** Stores data-relation names. */
   private ScopedSymbolDictionary drelDict = null;
   
   /** Maps keyword token types to the correct attribute and method type. */
   private Map<Integer, Integer> attributes = null;
   
   /** Provides access to ATTR_CLASS or METH_CLASS class names. */
   private Map<Integer, String> attrTypes = null;
   
   /** Current class or interface definition that is being created. */
   private Stack<ClassDefinition> currentCls = null;
   
   /** List of handles that could be possible active-x targets. */
   private Set<String> possibleActiveX = null;
   
   /** AST-specific variable definition index workaround. */
   private int tempIdx = 0;
   
   /** Current PROPATH. */
   private String[] propath = null;
   
   /** The legacy system's filesystem case-sensitivity. */
   private boolean caseSens = true;
   
   /** Used for recursive class or interface loading. */
   private AstGenerator generator = null;
   
   /** Flag to globally enable/disable table promotion. */
   private boolean allow = true;
   
   static
   {
      classDict = new ScopedSymbolDictionary(null, false);
      
      ACCESS_MODES.add(KW_PUBLIC);
      ACCESS_MODES.add(KW_PROTECTD);
      ACCESS_MODES.add(KW_PRIVATE);
   }
   
   /**
    * Default constructor that creates an instance with schema lookups
    * <b>disabled</b>.  Each namespace is initialized.
    */
   public SymbolResolver()
   {
      this(false, true, null);
   }
   
   /**
    * Constructs an instance with control over disabling schema lookups.
    * This allows an instance to be created in the absence of any schema
    * support.  Each namespace is initialized.
    *
    * @param    schema
    *           <code>true</code> enables schema dictionary support.
    */
   public SymbolResolver(boolean schema)
   {
      this(schema, true, null);
   }
   
   /**
    * Constructs an instance with control over disabling schema lookups.
    * This allows an instance to be created in the absence of any schema
    * support.  Each namespace is initialized.
    *
    * @param    schema
    *           <code>true</code> enables schema dictionary support.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    * @param    propath
    *           The current <code>PROPATH</code> to honor or <code>null</code>
    *           for the default of ".".
    */
   public SymbolResolver(boolean schema, boolean caseSens, String[] propath)
   {
      kwDict          = new KeywordDictionary();
      varDict         = new ScopedSymbolDictionary(null, false);
      funcDict        = new ScopedSymbolDictionary(null, false);
      procDict        = new ScopedSymbolDictionary(null, false);
      explPkgDict     = new LinkedHashMap<String, UsingSpec>();
      searchPkgDict   = new LinkedHashSet<UsingSpec>();
      labelDict       = new ScopedSymbolDictionary(null, false);      
      widgetDict      = new ScopedSymbolDictionary(null, false);
      frameDict       = new ScopedSymbolDictionary(null, false);
      menuDict        = new ScopedSymbolDictionary(null, false);
      menuItemDict    = new ScopedSymbolDictionary(null, false);
      streamDict      = new ScopedSymbolDictionary(null, false);
      dsetDict        = new ScopedSymbolDictionary(null, false);
      dsrcDict        = new ScopedSymbolDictionary(null, false);
      drelDict        = new ScopedSymbolDictionary(null, false);
      attributes      = new HashMap<Integer, Integer>();
      attrTypes       = new HashMap<Integer, String>();
      currentCls      = new Stack<ClassDefinition>();
      possibleActiveX = new HashSet<String>();
      
      if (schema)
      {
         // this load is dependent upon a configuration file being available
         try
         {
            schemaDict = new SchemaDictionary();
            
            // add a default scope that represents the scope of the external 
            // procedure
            addSchemaScope(false);
         }
         catch (SchemaException ex)
         {
            // in this case, the schema dictionary is invalid, however all
            // other types of lookups will work fine --> if there is no
            // schema name processing, then consuming this exception is safe
            // if there is schema name processing then get ready for a
            // null pointer exception!
            ex.printStackTrace();
         }
      }
      
      setPropath(propath);
      this.caseSens = caseSens;
      initPossibleClasses(this.propath, this.caseSens);
   }
   
   /**
    * Obtain the list of all "fakeout" class and interface definitions that
    * exist in the project.  These <code>.cls</code> files represent the
    * Progress built-ins and all .NET classes/interfaces that can be accessed.
    * <p>
    * These files will be listed in the directories represented by the
    * {@link #OO4GL_PATH}, {@link #DOTNET_PATH} and the {@link #ASSEMBLY_PATH}.
    *
    * @return   The list of <code>.cls</code> files, in the same case in which
    *           they exist in the filesystem.
    */
   public static String[] listFakeoutFiles()
   {
      String[] oo4gl    = getClassFileList(OO4GL_PATH, false);
      String[] dotnet   = getClassFileList(DOTNET_PATH, false);
      String[] assembly = getClassFileList(ASSEMBLY_PATH, false);
      
      ArrayList<String> list = new ArrayList<String>();
      
      list.addAll(Arrays.asList(oo4gl));
      list.addAll(Arrays.asList(dotnet));
      list.addAll(Arrays.asList(assembly));
      
      return  (String[]) list.toArray(new String[0]);
   }
   
   /**
    * Enables schema dictionary lookups when they are disabled at construction
    * and provides external control or sourcing over the schema dictionary.
    *
    * @param    schemaDict
    *           The schema dictionary to use.
    */
   public void setSchemaDictionary(SchemaDictionary schemaDict)
   {
      this.schemaDict = schemaDict;
   }
   
   /**
    * Gets the <code>PROPATH</code>.
    *
    * @return   The current <code>PROPATH</code> to honor.
    */
   public String[] getPropath()
   {
      return propath;
   }
   
   /**
    * Sets the <code>PROPATH</code>.
    *
    * @param    propath
    *           The current <code>PROPATH</code> to honor or <code>null</code>
    *           for the default of ".".
    */
   public void setPropath(String[] propath)
   {
      this.propath = (propath == null) ? new String[] { "." } : propath;
   }
   
   /**
    * Adds a <code>Keyword</code> to the keyword dictionary.
    *
    * @param   word
    *          <code>Keyword</code> to add.
    */
   public void addKeyword(Keyword word)
   {
      kwDict.addKeyword(word);
   }
   
   /**
    * Removes a <code>Keyword</code> from the keyword dictionary.
    *
    * @param   word
    *          <code>Keyword</code> to remove.
    */
   public void removeKeyword(Keyword word)
   {
      kwDict.deleteKeyword(word);
   }
   
   /**
    * Searches for a <code>Keyword</code> in the keyword dictionary, based
    * on a specific key.
    *
    * @param   key
    *          Text string to match.
    * @return  The <code>Keyword</code> that was found or null if no match
    *          was found.
    */   
   public Keyword lookupKeyword(String key)
   {
      return kwDict.lookupKeyword(key);
   }
   
   /**
    * Increases the scoping level of the variable and widget dictionaries by
    * 1 level.  All symbols must be unique within a scoping level.
    */   
   public void addScope()
   {
      varDict.addScope(null);
      widgetDict.addScope(null);
   }
   
   /**
    * Decreases the scoping level of the variable and widget dictionaries by
    * 1 level.  All symbols previously defined in the top-most scoping level
    * are lost and will no longer be found during lookups.
    */      
   public void deleteScope()
   {
      varDict.deleteScope();
      widgetDict.deleteScope();
   }
   
   /**
    * Persist schema dictionary information which is specific to the source
    * file currently being parsed.
    *
    * @param    srcFile
    *           Name of the 4GL source code file containing the schema
    *           definitions being persisted.
    * @param    schemaFile
    *           Name of schema persistence file.
    *
    * @throws   AstException
    *           if any error occurs writing persistence file.
    */
   public void persistSchemaData(String srcFile, String schemaFile)
   throws AstException
   {
      if (schemaDict != null)
      {
         schemaDict.persist(srcFile, schemaFile);
      }
   }
   
   /**
    * Adds a variable name and associated token type to the appropriate scope
    * of the variable dictionary, based on the content of the passed in
    * <code>Variable</code> wrapper object.
    *
    * @param    global
    *           Add to global scope if <code>true</code>.
    * @param    var
    *           Variable to add.
    */
   public void addVariable(boolean global, Variable var) 
   {
      addWrappedWorker(varDict,
                       global,
                       var.getName(),
                       (TokenDataWrapper) var);
   }   
   
   /**
    * Adds a widget name and associated token type to the appropriate scope
    * of the widget dictionary, based on the content of the passed in
    * <code>Variable</code> wrapper object.
    *
    * @param    global
    *           Add to global scope if <code>true</code>.
    * @param    wid
    *           Widget as represented in a <code>Variable</code> object.
    */
   public void addWidget(boolean global, Variable wid) 
   {
      addWrappedWorker(widgetDict,
                       global,
                       wid.getName(),
                       (TokenDataWrapper) wid);
   }   
   
   /**
    * Helper to load variable, widget and frame dictionaries based on a list
    * of <code>Variable</code> objects. This processing will load all 3
    * dictionaries from the same list, however it must be noted that adding
    * a variable does NOT result in a matching widget of the same name in
    * the widget dictionary (even though this is essentially what happens
    * when the parser adds new variables to the variable dictionary). To 
    * acheive this same effect, one must add multiple <code>Variable</code>
    * objects to the list as needed.
    * <p>
    * A field can be added to this lookup by using a field token type instead
    * of a variable token type. This will allow an ambiguous field reference
    * to be made unambiguous since variable lookups are processed
    * <b>before</b> any schema dictionary lookups.
    *
    * @param    varlist
    *           List of variables, widgets, frames or fields to add to the
    *           appropriate dictionary.
    */
   public void loadVariableSymbols(Variable[] varlist)
   {
      for ( int j = 0; j < varlist.length; j++ )
      {
         int type = varlist[j].getTokenType();
         
         if (
              (type >= ProgressParser.BEGIN_VARTYPES   &&
               type <= ProgressParser.END_VARTYPES)    ||
              (type >= ProgressParser.BEGIN_FIELDTYPES &&
               type <= ProgressParser.END_FIELDTYPES)
            )
         {
            addVariable(true, varlist[j]);
         }
         else if (type == ProgressParser.WID_FRAME)
         {
            addFrame(varlist[j].getName(), ProgressParser.WID_FRAME);
         }         
         else if (type >= ProgressParser.BEGIN_WIDGETS &&
                  type <= ProgressParser.END_WIDGETS )
         {
            addWidget(true, varlist[j]);
         }         
      }
   }
   
   /**
    * Adds a variable name and associated token type to the global scope
    * of the variable dictionary.  Names in this scope can be hidden by
    * duplicate names added to scopes that are closer to the top of the
    * stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    tokenType
    *           Lexer and Parser token type associated with the name. 
    */
   public void addGlobalVariable(String name, int tokenType) 
   {
      addGlobalVariable(name, tokenType, null);
   }
   
   /**
    * Adds a variable name and associated token type to the global scope
    * of the variable dictionary.  Names in this scope can be hidden by
    * duplicate names added to scopes that are closer to the top of the
    * stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    tokenType
    *           Lexer and Parser token type associated with the name. 
    * @param    cls
    *           Fully qualified class name of the object instance represented
    *           or <code>null</code> if this variable does not represent an
    *           object instance.
    */
   public void addGlobalVariable(String name, int tokenType, String cls) 
   {
      addWrappedWorker(varDict,
                       true,
                       name,
                       (TokenDataWrapper) new Variable(name, tokenType, cls));
   }
   
   /**
    * Adds a variable name and associated token type to the current scope
    * of the variable dictionary.  This is the same as the top-most scope.
    * Names in this scope can hide duplicate names added to scopes that are
    * lower on the stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    tokenType
    *           Lexer and Parser token type associated with the name. 
    * @param    cls
    *           Fully qualified class name of the object instance represented
    *           or <code>null</code> if this variable does not represent an
    *           object instance.
    */   
   public void addLocalVariable(String name, int tokenType, String cls) 
   {
      addWrappedWorker(varDict,
                       false,
                       name,
                       (TokenDataWrapper) new Variable(name, tokenType, cls));
   }

   /**
    * Adds a variable name and associated token type/options to the current
    * scope of the variable dictionary based on a referenced variable or
    * schema field in a <code>LIKE</code> clause.  This is the same as the
    * top-most scope.  Names in this scope can hide duplicate names added to
    * scopes that are lower on the stack of scopes.
    *
    * @param    name
    *           Variable name to add.
    * @param    ast
    *           AST node containing the like clause. 
    */   
   public void addLocalVariableLike(String name, Aast ast) 
   { 
      int     ltype   = ast.getType();
      int     type    = ltype;
      String  oldname = ast.getText();
      int     extent  = getExtentSubscript(ast);
      boolean isFld   = ltype > BEGIN_FIELDTYPES && ltype < END_FIELDTYPES;
      String  cls     = null;
   
      // the local variable type must be converted to the proper
      // VAR_ variant *if* this was a like FIELD_...  
      switch (ltype)
      {
         case FIELD_CHAR:
            type = VAR_CHAR;
            break;
         case FIELD_CLASS:
            type = VAR_CLASS;
            // class fields must propagate the fully qualified class name
            cls  = (String) ast.getAnnotation("qualified");
            break;
         case FIELD_COM_HANDLE:
            type = VAR_COM_HANDLE;
            break;
         case FIELD_DATE:
            type = VAR_DATE;
            break;
         case FIELD_DATETIME:
            type = VAR_DATETIME;
            break;
         case FIELD_DATETIME_TZ:
            type = VAR_DATETIME_TZ;
            break;
         case FIELD_DEC:
            type = VAR_DEC;
            break;
         case FIELD_HANDLE:
            type = VAR_HANDLE;
            break;
         case FIELD_INT:
            type = VAR_INT;
            break;
         case FIELD_INT64:
            type = VAR_INT64;
            break;
         case FIELD_LOGICAL:
            type = VAR_LOGICAL;
            break;
         case FIELD_RAW:
            type = VAR_RAW;
            break;                     
         case FIELD_RECID:
            type = VAR_RECID;
            break;
         case FIELD_ROWID:
            type = VAR_ROWID;
            break;
         default:
            // this MUST be at the end of the list of fields, it will match
            // BLOB and CLOB which can't be variable types
            if (isFld)
            {
               String err = String.format("Unexpected field type %s!",
                                          ast.getDescriptiveTokenText());
               throw new RuntimeException(err);
            }
            break;
      }
      
      // class variables must propagate the fully qualified class name 
      if (ltype == VAR_CLASS)
         cls = (String) ast.getAnnotation("qualified");
      
      Variable var   = new Variable(name, type, cls);
      Aast     fopts = null;
      
      if (isFld)
      {
         // we must pattern the new variable off a referenced field         
         // obtain the field options AST from the schema
         try
         {
            fopts = schemaDict.getFieldCopy(oldname, extent);
         }
         catch (AmbiguousSchemaNameException exc)
         {
            exc.printStackTrace();
         }
      }
      else
      {
         // we must pattern the variable off another variable
         Variable like = (Variable) lookupWrapped(varDict, oldname);
         fopts = like.getOptions(extent);
      }
      
      if (fopts != null)
      {                                           
         var.copyDefaultOptions(fopts);
      }
      
      addWrappedWorker(varDict, false, name, (TokenDataWrapper) var);
   }
   
   
   /**
    * Inspect the variable definition and write any non-default options into 
    * the variable object.
    *
    * @param    name
    *           The variable name which must have its options set.
    * @param    ast
    *           The tree to inspect which defines a Progress variable.
    */
   public void setVariableOptions(String name, Aast ast)
   {
      Variable var = (Variable) lookupWrapped(varDict, name);
      
      // was the variable object found?
      if (var == null)
      {
         // this may be a widget that has no var backer (e.g. a button)
         var = (Variable) lookupWrapped(widgetDict, name);
      }      
      
      if (var == null || ast == null)
      {
         // nothing to do
         return;
      }
      
      // the KW_LIKE should already have been processed by now (in the 
      // addLocalVariableLike() so that defaults are set from that       
      // variable/field and then can be overridden by explicit options      
      // so now we process the explicit overrides
      
      var.setOptions(ast);
   }
   
   /**
    * Stores all non-standard options as annotations to the AST passed as
    * a parameter if this is the original definition, otherwise it stores
    * a subset of values, most important of which is the cross-reference
    * index to allow the original definition to be easily identified and
    * accessed from a reference.  If an option has the default value, it is
    * not saved as an annotation.
    *
    * @param    name
    *           The variable name which must have its options annotated.
    * @param    ast
    *           The AST to annotate.    
    * @param    original
    *           <code>true</code> if this is to annotate the AST of the
    *           original definition, <code>false</code> if the AST is only
    *           a reference.
    */
   public void annotateVariableOptions(String  name,
                                       Aast    ast,
                                       boolean original)
   {
      // assume this is a variable
      Variable var = (Variable) lookupWrapped(varDict, name);
      
      // was the variable object found?
      if (var == null)
      {
         // this may be a widget that has no var backer (e.g. a button)
         var = (Variable) lookupWrapped(widgetDict, name);
      }
      
      // if called from the ProgressExpressionEvaluator (to evaluate an
      // expression) the normal variable dictionary processing is bypassed
      // which means this can be null, just return in this case OTHERWISE
      // we annotate
      if (var != null)
         var.annotateOptions(ast, original, this);
   }
   
   /**
    * Creates a new field AST node based on a qualified or unqualified
    * field name (must be unambiguous) and sets the token type, text and
    * all annotations as needed.  This centralizes a service which is
    * normally handled by the parser directly.  It uses the
    * {@link #annotateField} method for the core of the lookup and
    * annotation logic.
    * <p>
    * No table promotion will occur based on this field name, even if it
    * is a qualified name.
    *
    * @param    name
    *           The field name from which to obtain options and type.
    *
    * @return   The created AST node.
    */
   public Aast createField(String name)
   {
      ProgressAst field = new ProgressAst();
      
      field.setText(name);
      annotateField(field, true, false);
      
      return field;
   }
   
   /**
    * Stores common data (schemaname, bufname, recordtype) as well as
    * all non-standard options as annotations to the AST passed as
    * a parameter.  If an option has the default value, it is not saved as
    * an annotation.  Optionally sets the type of the given AST based on
    * the field found in the schema dictionary.  The field will be looked
    * up in the dictionary based on the text of the AST node and there
    * must be no ambiguity.
    * <p>
    * If the field cannot be found due to ambiguity (more than 1 match), the
    * method will return with no resulting changes.
    * <p>
    * If no such field exists, some default annotations will be set to
    * "unknown".
    *
    * @param    field
    *           The AST to annotate.    
    * @param    setType
    *           If <code>true</code>, the AST node will have its type set
    *           based on the field found in the schema dictionary.
    * @param    promote
    *           <code>true</code> to force the promotion (in the schema
    *           dictionary) of the table if the field's name is qualified.
    */
   public void annotateField(Aast field, boolean setType, boolean promote)
   {      
      String ftext = field.getText();
      
      Aast fback = null;
               
      // obtain the backing field AST from the schema
      try
      {
         fback = schemaDict.getField(ftext);
      }
      catch (AmbiguousSchemaNameException exc)
      {
         exc.printStackTrace();
         return;
      }
      
      // make sure we found the backing field
      if (fback != null)
      {
         if (setType)
         {
            field.setType(fback.getType());
         }
         
         // now store an annotation with the fully qualified name and
         // the shorted unabbreviated unique buffer name               
         String fullname = lookupFullFieldName(ftext);
         String bufname  = lookupBufferName(ftext, true);
         String dbname = lookupDatabaseName(bufname);
         
         // was the original reference qualified?
         if (promote && ftext.indexOf('.') != -1)
         {
            // qualified references cause table promotion
            promoteTableName(bufname, false, false);
         }
         
         if (fullname != null)
            field.putAnnotation("schemaname", fullname);
            
         if (bufname != null)
            field.putAnnotation("bufname", bufname);
         
         if (dbname != null)
            field.putAnnotation("dbname", bufname);
            
         // store the backing construct (this will be TABLE, TEMP_TABLE
         // or WORK_TABLE)
         int backing = lookupFieldRecordType(ftext);
         
         if (backing != -1)         
            field.putAnnotation("recordtype", new Long(backing));
         
         int    newtype = field.getType();
         String cls     = (newtype == FIELD_CLASS) ? ROOT_OBJ_NAME : null; 
         
         // temporary variable object to be our worker to annotate
         Variable var = new Variable(ftext, newtype, cls);
         
         // we must pattern the new variable off a referenced field
         var.copyDefaultOptions(fback);
         var.annotateOptions(field, true, this);
      }
      else
      {
         // if the FIELD_ type was set via a hint or some other
         // unexpected manner, then this will get triggered...
         field.putAnnotation("schemaname", "unknown field");
         field.putAnnotation("bufname", "unknown buffer");
         field.putAnnotation("recordtype", new Long(-1));
      }
   }
   
   /**
    * Searches the variable dictionary for a symbol matching the passed 
    * key or matching a keyword's full text if the symbol matches a language
    * keyword.  This lookup starts at the top-most scoping level and descends
    * through each scope level on the stack until a match is found. The
    * token type of the first match found is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   actual
    *          Actual text string to match which may be a language keyword
    *          abbreviation.
    *
    * @return  The token type that was found or -1 if no match was found.
    */    
   public int lookupVariable(String actual)
   {
      return lookupWrappedType(lookupWrapped(varDict, actual));
   }
      
   /**
    * Searches a dictionary containing {@link TokenDataWrapper} objects for a
    * symbol matching the passed key or matching a keyword's full text if the
    * symbol matches a language keyword.  This lookup starts at the top-most
    * scoping level and descends through each scope level on the stack until
    * a match is found. The wrapped object of the first match found is
    * returned.
    *
    * @param   ssd
    *          The dictionary to search.
    * @param   actual
    *          Actual text string to match which may be a language keyword
    *          abbreviation.
    *
    * @return  The wrapped object that was found or <code>null</code> if no
    *          match was found.
    */    
   public TokenDataWrapper lookupWrapped(ScopedSymbolDictionary ssd,
                                         String                 actual)
   {
      Keyword          kw      = lookupKeyword(actual);
      TokenDataWrapper wrapper = null;
      
      if (kw != null)
      {
         // the actual text matches a keyword
         if (kw.isReserved())
         {
            // since this is a reserved keyword, this must be a 
            // built-in variable/function
            wrapper = (TokenDataWrapper) ssd.lookupSymbol(kw.getFullText());
         }
         else
         {
            // this is an unreserved keyword, this MAY be a 
            // built-in variable/function BUT we must allow a user-defined
            // name to take precedence (note that if the actual text is
            // is the same as the full keyword text, then we may still
            // get a match here that is the built-in, assuming no
            // user-defined name exists --> that is OK too)
            wrapper = (TokenDataWrapper) ssd.lookupSymbol(actual);
            
            // check if there was a match to the actual text
            if (wrapper == null)
            {
               // match the full text of the keyword with the built-in
               wrapper = (TokenDataWrapper) ssd.lookupSymbol(kw.getFullText());
            }
         }
      }
      else
      {
         // there is no keyword, so this must be a user-defined name
         wrapper = (TokenDataWrapper) ssd.lookupSymbol(actual);
      }            
      
      return wrapper;
   }   
      
   /**
    * Adds a user-defined function and its return type to the dictionary
    * of functions. This is a flat dictionary (there is no scoping).  The 
    * function name is used as the key and a wrapped {@link Function} object
    * is the value.  Since the dictionary is flat, all function names must be
    * unique.
    * <p>
    * Due to the Progress forward definition capability, this function only
    * adds to the dictionary the first time this method is called for a
    * given name.
    *
    * @param    name
    *           Function name.
    * @param    tokenType
    *           The token type of the function's return value as defined by
    *           the Lexer and Parser.
    * @param    qname
    *           Fully qualified class name for functions that return an
    *           object instance, otherwise <code>null</code>.
    */
   public void addFunction(String name, int tokenType, String qname) 
   {
      // a forward declaration may have already occurred for this function
      // try a lookup
      Function func = (Function) funcDict.lookupSymbol(name);
      
      // is this a new function def OR an override of a builtin function?
      if (func == null || func.isBuiltin())
      {
         func = new Function(name, tokenType, false, qname); 
         
         addWrappedWorker(funcDict,
                          false,
                          name, 
                          (TokenDataWrapper) func);
      }
   }

   /**
    * Adds a built-in Progress function and its return type to the dictionary
    * of functions. This is a flat dictionary (there is no scoping).  The 
    * function name is used as the key and a wrapped {@link Function} object
    * is the value.  Since the dictionary is flat, all function names must be
    * unique.
    *
    * @param    name
    *           Function name.
    * @param    tokenType
    *           The token type of the function's return value as defined by
    *           the Lexer and Parser.
    * @param    returnsUnknown
    *           Specifies if this built-in function can possibly return
    *           the unknown value.
    * @param    qname
    *           Fully qualified class name for functions that return an
    *           object instance, otherwise <code>null</code>.
    */
   public void addBuiltinFunction(String  name,
                                  int     tokenType,
                                  boolean returnsUnknown,
                                  String  qname) 
   {
      Function func = new Function(name, tokenType, true, qname);
      func.setReturnsUnknown(returnsUnknown ? Boolean.TRUE : Boolean.FALSE);
      
      addWrappedWorker(funcDict,
                       false,
                       name, 
                       (TokenDataWrapper) func);
   }

   /**
    * Adds a built-in Progress function and its return type to the dictionary
    * of functions. This is a flat dictionary (there is no scoping).  The 
    * function name is used as the key and a wrapped {@link Function} object
    * is the value.  Since the dictionary is flat, all function names must be
    * unique.
    *
    * @param   name
    *          Function name.
    * @param   tokenType
    *          The token type of the function's return value as defined by
    *          the Lexer and Parser.
    * @param   returnsUnknown
    *          Specifies if this built-in function can possibly return
    *          the unknown value.
    */
   public void addBuiltinFunction(String  name,
                                  int     tokenType,
                                  boolean returnsUnknown) 
   {
      addBuiltinFunction(name, tokenType, returnsUnknown, null);
   }

   /**
    * Searches the function dictionary for a symbol matching the passed 
    * key or matching a keyword's full text if the symbol matches a language
    * keyword.  This lookup starts at the top-most scoping level and descends
    * through each scope level on the stack until a match is found. The
    * token type of the first match found is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param    actual
    *           Actual text string to match which may be a language keyword
    *           abbreviation.
    *
    * @return   The token type that was found or -1 if no match was found.
    */    
   public int lookupFunction(String actual)
   {
      return lookupWrappedType(lookupWrapped(funcDict, actual));
   }         
   
   /**
    * Searches the function dictionary for a symbol matching the passed 
    * key or matching a keyword's full text if the symbol matches a language
    * keyword.  This lookup starts at the top-most scoping level and descends
    * through each scope level on the stack until a match is found. The
    * definition of the first match found is returned.
    *
    * @param    actual
    *           Actual text string to match which may be a language keyword
    *           abbreviation.
    *
    * @return   The function definition or <code>null</code> if no match was
    *           found.
    */    
   public Function lookupFunctionDef(String actual)
   {
      return (Function) lookupWrapped(funcDict, actual);
   }         
   
   /**
    * Stores all non-standard options as annotations to the AST passed as
    * a parameter if this is the original definition, otherwise it stores
    * a subset of values, most important of which is the cross-reference
    * index to allow the original definition to be easily identified and
    * accessed from a reference.  If an option has the default value, it is
    * not saved as an annotation.
    *
    * @param    name
    *           The function name which must have its options annotated.
    * @param    ast
    *           The AST to annotate.    
    * @param    original
    *           <code>true</code> if this is to annotate the AST of the
    *           original definition, <code>false</code> if the AST is only
    *           a reference.
    */
   public void annotateFunction(String name, Aast ast, boolean original)
   {
      Function func = (Function) lookupWrapped(funcDict, name);
      
      // special case: preprocessor expression evaluation can resolve to
      // function token types without any function being registered since
      // the CallbackResolver registers the function token types inside
      // the KeywordDictionary; this means that a null pointer can occur
      // here, so we just return without doing any annotations in that
      // case
      if (func == null)
         return;
      
      func.annotateOptions(ast, original, this);
   }
      
   /**
    * Adds a procedure and a token type to the dictionary of procedures.
    * This is a flat dictionary (there is no scoping).  The procedure name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all procedure names must be unique.
    *
    * @param   name
    *          Procedure name.
    * @param   tokenType
    *          The procedure's token type as defined by the Lexer and Parser.
    */   
   public void addProcedure(String name, int tokenType) 
   {
      procDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the procedure dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The procedure's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupProcedure(String name)
   {
      return lookupWorkerExact(procDict, name);
   }   
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    symbol
    *           Node defining the fully qualified (including package) class
    *           name.
    * @param    inherits
    *           Node defining the fully qualified name of the superclass or
    *           <code>null</code> for none.
    * @param    imple
    *           Node which has the list of interfaces that are implemented
    *           or <code>null</code> if none exist.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    dotnet
    *           <code>true</code> if this definition is a .NET class.
    */   
   public void addClass(Aast    symbol,
                        Aast    inherits,
                        Aast    imple,
                        boolean builtin,
                        boolean dotnet) 
   {
      ClassDefinition parent = null;
      Set<String>     ifaces = null;
      String          name   = symbol.getText();
      
      if (inherits != null)
      {
         Aast   child = inherits.getChildAt(0);
         String pname = (String) child.getAnnotation("qualified");
         parent = lookupClass(pname);
         
         if (parent == null)
         {
            // will cause recursion if successful
            loadClass(pname);
            parent = lookupClass(pname);
         }
            
         if (parent == null)
         {
            String err = String.format("Missing class definition for %s!",
                                       pname);
            throw new RuntimeException(err);
         }
      }
      else
      {
         // default parent for all classes
         parent = lookupClass(ROOT_OBJ_NAME);
      }
      
      if (imple != null)
      {
         ifaces = new HashSet<String>();
         
         Aast child = imple.getChildAt(0);
         
         while (child != null)
         {
            ifaces.add((String) child.getAnnotation("qualified"));
            child = (Aast) child.getNextSibling();
         }
      }
      
      addClass(name, parent, ifaces, builtin, dotnet);
   }
   
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    name
    *           The fully qualified (including package) class name.
    * @param    parent
    *           The superclass or <code>null</code> for none.
    * @param    ifaces
    *           The list of interfaces that are implemented or
    *           <code>null</code> if none exist.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    dotnet
    *           <code>true</code> if this definition is a .NET class.
    */   
   public void addClass(String          name,
                        ClassDefinition parent,
                        Set<String>     ifaces,
                        boolean         builtin,
                        boolean         dotnet) 
   {
      ClassDefinition cls = new ClassDefinition(name,
                                                parent,
                                                ifaces,
                                                builtin,
                                                false,
                                                dotnet);
                                                
      // remember this is the class definition being updated (with methods,
      // members and properties)
      currentCls.push(cls);
      
      synchronized (SymbolResolver.class)
      {
         classDict.addSymbol(false, name, cls);
      }
   }
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    symbol
    *           Node defining the fully qualified (including package) class
    *           name.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    dotnet
    *           <code>true</code> if this definition is a .NET class.
    */
   public void addInterface(Aast symbol, boolean builtin, boolean dotnet) 
   {
      addInterface(symbol.getText(), builtin, dotnet);
   }
   
   /**
    * Adds a class name and a token type to the dictionary of classes.
    * This is a flat dictionary (there is no scoping).  The class name
    * is used as the key and the token type is the value.  Since the
    * dictionary is flat, all class names must be unique.
    *
    * @param    name
    *           The fully qualified (including package) class name.
    * @param    builtin
    *           <code>true</code> if this definition is built-in to Progress.
    * @param    dotnet
    *           <code>true</code> if this definition is a .NET class.
    */   
   public void addInterface(String name, boolean builtin, boolean dotnet) 
   {
      ClassDefinition iface = new ClassDefinition(name,
                                                  null,
                                                  null,
                                                  builtin,
                                                  true,
                                                  dotnet);
                                                  
      // remember this is the interface definition being updated (with
      // methods, members and properties)
      currentCls.push(iface);
      
      synchronized (SymbolResolver.class)
      {
         classDict.addSymbol(false, name, iface);
      }
   }
   
   /**
    * Obtain the currently active class or interface definition.
    *
    * @return   The active class or interface definition or <code>null</code>
    *           if there is no active definition.
    */
   public ClassDefinition getCurrentClassDef()
   {
      return currentCls.empty() ? null : currentCls.peek();
   }
   
   /**
    * Obtain the name of the currently active class definition.
    *
    * @return   The class name or <code>null</code> if there is no active
    *           class definition.
    */
   public String getCurrentClassName()
   {
      ClassDefinition cls = getCurrentClassDef();
      return (cls == null) ? null : cls.getName();
   }
   
   /**
    * Obtain the parent of the currently active class definition.
    *
    * @return   The parent class definition or <code>null</code> if there is
    *           no explicit parent.
    */
   public ClassDefinition getCurrentParentDef()
   {
      ClassDefinition cls = getCurrentClassDef();
      return (cls == null) ? null : cls.getParent();
   }
   
   /**
    * Obtain the name of the parent of the currently active class definition.
    *
    * @return   The parent class name.
    */
   public String getCurrentParentName()
   {
      ClassDefinition cls = getCurrentParentDef();
      return (cls == null) ? ROOT_OBJ_NAME : cls.getName();
   }
   
   /**
    * Notify the symbol resolver that the current class or interface
    * definition is complete.
    */
   public void classDefinitionComplete()
   {
      currentCls.pop();
   }
   
   /**
    * Attempt to load the class or interface name given. This "eats" any
    * failures since it is not guaranteed that the input does represent a
    * valid class name.
    *
    * @param    name
    *           A potential class or interface name.
    *
    * @return   <code>true</code> if the class or interface can be loaded.
    */
   public boolean canLoadClass(String name)
   {
      String full = null;
      
      try
      {
         full = loadClass(name);
      }
      
      catch (Exception exp)
      {
         // ignore
      }
      
      return (full != null);
   }
   
   /**
    * Load (using recursion if needed) the hierarchy of class definitions (or
    * a single interface definition) referenced by the given name. This should
    * be called when a reference to a class or interface name occurs (except
    * in the case of the <code>CLASS</code> or <code>INTERFACE</code> language
    * statements). The name can be unqualified (requiring a call to
    * {@link #resolveClassName} to determine the fully qualified name) or it
    * can be fully qualified. The name cannot (just like in Progress) be
    * partially qualified. Once resolved, if the class or interface definition
    * is already loaded, then this will return immediately.  Otherwise the
    * class or interface file will be found via the <code>PROPATH</code> and
    * that file's AST will be loaded.  If the AST does not exist, then the
    * file will first be parsed. During parsing, the <code>CLASS</code> or 
    * <code>INTERFACE</code> language statements will call {@link #addClass}
    * or {@link #addInterface} respectively. This will instantiate the proper
    * {@link ClassDefinition} objects and the contained definitions for
    * methods, member variables and properties will be added by the rules
    * that match those language features. Once parsing is complete, the
    * definition will be fully formed and loaded. Any references to other
    * classes or interfaces in that file will cause recursive parsing/loading
    * using this method, until the entire object hierarchy is loaded. At that
    * point, this chain of methods will return. This recursive loading can
    * also be duplicated by tree walking for an AST that already exists. This
    * processing is the same as that done by parsing but it just avoids the
    * overhead of the full parsing. However, it is triggered by the same
    * language features as noted above. This method uses parsing and/or tree
    * walking as needed and both these techniques may be intermixed in the
    * same recursive chain of class/interface loading.
    *
    * @param    name
    *           Unqualified or fully qualified class or interface name to
    *           load.
    *
    * @return   The fully qualified class or interface name that was loaded.
    */
   public String loadClass(String name)
   {
      SearchResult found = resolveClassName(name);
      
      ClassDefinition cls = lookupClass(found.clsname);
      
      if (cls == null)
      {
         // not loaded yet
         
         // do we have a cached file name?
         if (found.filename == null)
         {
            // find the class file name in the propath
            if (!findFile(found.clsname, found.type, found))
            {
               String spec = "Cannot find class/interface %s in PROPATH.";
               String err  = String.format(spec, found.clsname);
               throw new RuntimeException(err);
            }
            
            // at this point found.filename must be valid
         }
         
         if (generator == null)
         {
            generator = new AstGenerator();
            generator.setPreprocess(true);
            generator.setCache(true);
            generator.setDumpLexer(true);
            generator.setDumpParser(true);
            generator.setAstPersist(true);
         }
         
         String relative = Configuration.normalizeFilename(found.filename);
         
         Aast ast = null;
         
         try
         {
            // parse or load an already parsed AST (if we parse here, the
            // parser may call this method again, but in a separate symbol
            // resolver)
            ast = generator.processFile(relative,
                                        null,
                                        -1,
                                        found.builtin,
                                        found.dotnet);
         }
         
         catch (AstException ae)
         {
            ae.printStackTrace();
         }
         
         // if we had to parse instead of load, then the class definition
         // is already there
         cls = lookupClass(found.clsname);
         
         if (cls == null)
         {
            // we loaded the AST since the file was already parsed
            // previously, so now we must walk the tree and load the
            // class/interface definitions the same as would have happened
            // during parsing (may indirectly recurse into this method)
            loadViaWalk(ast);
            
            if (lookupClass(found.clsname) == null)
            {
               String spec = "Cannot load class/interface %s from file %s!";
               String err  = String.format(spec, found.clsname, relative);
               throw new RuntimeException(err);
            }
         }
      }
      
      return found.clsname;
   }

   /**
    * Searches the class dictionary for a definition matching the passed 
    * key.
    *
    * @param   name
    *          Fully qualified class/interface name to match.
    *
    * @return  The class that was found or <code>null</code> if no match was
    *          found.
    */             
   public ClassDefinition lookupClass(String name)
   {
      ClassDefinition cls = null;
      
      synchronized (SymbolResolver.class)
      {
         cls = (ClassDefinition) classDict.lookupSymbol(name);
      }

      return cls;
   }   
   
   /**
    * Annotate the given node which is a reference to the named class.
    *
    * @param   qname
    *          Fully qualified class name being referenced.
    * @param   node
    *          The node to annotate.
    */             
   public void annotateClassRef(String qname, Aast node)
   {
      ClassDefinition cls = null;
      
      synchronized (SymbolResolver.class)
      {
         cls = (ClassDefinition) classDict.lookupSymbol(qname);
      }

      node.putAnnotation("qualified", qname);
      node.putAnnotation("builtin", cls.isBuiltIn());
      node.putAnnotation("dotnet", cls.isDotNet());
   }   
   
   /**
    * Converts any unqualified class or interface names into the fully
    * qualified name (which has a package included). There is no such thing
    * as a partially qualified class or interface name in Progress, so if
    * any "." character is found in the source name, that name is assumed to
    * be fully qualified and is immediately returned. Otherwise, the
    * following name resolution actions are attempted:
    * <p>
    * <ol>
    *    <li> If the name matches the unqualified name of any of the
    *         explicitly imported classes (via the USING statement), that
    *         fully qualified class name is returned.
    *    <li> The "*" in each of the imported package specifications is
    *         replaced with the unqualified name (in order) and:
    *    <ol>
    *       <li> If the resulting name matches a class that is already loaded,
    *            that class name is set into the search result, but the other
    *            search result members will NOT be set.
    *       <li> The following paths are searched depending on the USING type:
    *          <ul>
    *             <li> <code>PROPATH</code> for the project's 4GL cls files
    *             <li> <code>OO4GL_PATH</code> for built-in 4GL cls files
    *             <li> <code>ASSEMBLY_PATH</code> for the project's .NET cls
    *                  files
    *             <li> <code>DOTNET_PATH</code> for built-in .NET cls files
    *          </ul>
    *       <li> If a match is found via the search, the members of the search
    *            result will be set as noted below.
    *    </ol>
    * </ol>
    * <p>
    * If an unqualified name is not matched to the above criteria, the name
    * is assumed to be the complete type name and is returned unchanged. This
    * means that no search has been successful.
    * <p>
    * If the filesystem search has been successful, the returned search
    * result will have a non-null <code>filename</code> member.  In addition,
    * the <code>builtin</code> and <code>dotnet</code> members will have been
    * set and can be relied upon.
    *
    * @param    name
    *           Unqualified or fully qualified class or interface name.
    *
    * @return   An object containing data about the search. If the 
    *           <code>filename</code> member is <code>null</code> then only
    *           the <code>name</code> member has been set, though it may not
    *           be valid. If the <code>filename</code> member is NOT
    *           <code>null</code>, then all members of the object are already
    *           set and can be relied upon.
    */             
   public SearchResult resolveClassName(String name)
   {
      ClassDefinition cls       = null;
      String          qualified = null;
      UsingSpec       uspec     = null;
      UsingType       utype     = UsingType.UNSPECIFIED;
      SearchResult    result    = new SearchResult();
      
      // remove array brackets if they exist (this assumes they are always
      // at the end AND that there is no other valid content after the
      // opening square bracket); note that brackets can define multi-
      // dimensional arrays like MyClass[,,] for a 3-dimensional array
      int sqBrackets = name.indexOf("[");
      
      if (sqBrackets >= 0)
      {
         name = name.substring(0, sqBrackets);
      }
      
      if (name.indexOf('.') == -1)
      {
         // check the explicit list of class names
         uspec     = explPkgDict.get(name.toLowerCase());
         qualified = (uspec != null) ? uspec.spec : null;
         
         if (qualified == null)
         {
            Iterator<UsingSpec> iter = searchPkgDict.iterator();
            
            // try each of the search specifications in order
            while (iter.hasNext() && qualified == null)
            {
               uspec = iter.next();
               
               // build the next possible qualified class name
               String next = String.format("%s%s", uspec.spec, name);
               
               // short circuit the search logic in the case that we already
               // have loaded this class definition
               cls = lookupClass(next);
               
               if (cls != null)
               {
                  qualified = next;
                  result.builtin = cls.isBuiltIn();
                  result.dotnet  = cls.isDotNet();
               }
               else
               {
                  // search propath for the specific class file name
                  if (findFile(next, uspec.type, result))
                  {
                     qualified = next;
                  }
               }
            }
         }
         
         if (qualified == null)
         {
            // the unqualified name must be the full name
            qualified = name;
         }
      }
      else
      {
         // the unqualified name must be the full name
         qualified = name;
      }
      
      if (uspec != null)
         utype = uspec.type;
      
      result.clsname  = qualified;
      result.type     = utype;
      
      return result;
   }   
   
   /**
    * Adds a package search specification ("package.*") or a fully qualified
    * class name ("package.class") to the list of known packages. These are
    * used to match unqualified class names with their fully qualified
    * equivalents.
    *
    * @param   name
    *          Package search specification or explicit fully qualified
    *          class name.
    * @param   fromClause
    *          The nodes parsed as the <code>FROM (ASSEMBLY | PROPATH)</code>
    *          clause or <code>null</code> if it wasn't present.
    */   
   public void addPackage(String name, Aast fromClause) 
   {
      UsingType type = UsingType.UNSPECIFIED;
      
      if (fromClause != null)
      {
         Aast child = fromClause.getChildAt(0);
         
         if (child != null)
         {
            type = (child.getType() == KW_ASSEMBLY) ? UsingType.DOTNET
                                                    : UsingType.PROPATH;
         }
      }
      
      if (name.endsWith(".*"))
      {
         // remove the asterisk since we will no longer need it (it only
         // serves to denote that this is a search spec versus an explicit
         // fully qualified class name 
         name = name.substring(0, name.length() - 1);
         
         // error out if the directory can't be found in the PROPATH
         if (!pkgExists(name))
         {
            String err = String.format("Non-existent package %s!", name);
            throw new RuntimeException(err);
         }
         
         searchPkgDict.add(new UsingSpec(name, type));
      }
      else
      {
         // find the base name
         int idx = name.lastIndexOf(".");
         
         if (idx < 0)
         {
            throw new RuntimeException("Invalid fully qualified class name!");
         }
         
         String base = name.substring(idx + 1);
         
         // map the lowercased base name to the fully qualified name (allows
         // case-insensitive lookup)
         explPkgDict.put(base.toLowerCase(), new UsingSpec(name, type));
      }
   }
   
   /**
    * Add the named method to the currently active class definition.
    *
    * @param    name
    *           Method name.
    * @param    mtype
    *           Return type for the method.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic                          
    *           <code>true</code> if the variable/property is static.
    * @param    node
    *           The AST to be annotated.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addObjectMethod(String  name,
                               int     mtype, 
                               int     access,
                               boolean isStatic,
                               Aast    node,
                               String  qname)
   {
      // must be non-null since methods can't be defined outside of classes
      currentCls.peek().addMethod(name, mtype, access, isStatic, node, qname);
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    access
    *           Node of type <code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>.
    * @param    stic
    *           Node of type <code>KW_STATIC</code> if the variable/property is
    *           static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String name,
                             Aast   access,
                             Aast   stic,
                             String qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         addDataMember(name, lookupVariable(name), access, stic, qname);
      }
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    type
    *           Token type of the member.
    * @param    access
    *           Node of type <code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>.
    * @param    stic
    *           Node of type <code>KW_STATIC</code> if the variable/property is
    *           static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String name,
                             int    type,
                             Aast   access,
                             Aast   stic,
                             String qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         int dmode = (type == DEFINE_PROPERTY) ? KW_PUBLIC : KW_PRIVATE;
         int amode = (access == null) ? KW_PUBLIC : access.getType();
                                      
         addDataMember(name, type, amode, (stic != null), qname);
      }
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic                          
    *           <code>true</code> if the variable/property is static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */                                                          
   public void addDataMember(String  name,
                             int     access,
                             boolean isStatic,
                             String  qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         addDataMember(name, lookupVariable(name), access, isStatic, qname);
      }
   }
   
   /**
    * Add the named variable as a data member of the currently active class
    * definition. If no class definition is active, this method does nothing.
    *
    * @param    name
    *           Method name.
    * @param    type
    *           Token type for the member.
    * @param    access
    *           Access mode (<code>KW_PUBLIC</code>, <code>KW_PROTECTD</code>
    *           or <code>KW_PRIVATE</code>).
    * @param    isStatic                          
    *           <code>true</code> if the variable/property is static.
    * @param    qname
    *           Fully qualified class name for <code>mtype</code> methods of
    *           <code>METH_CLASS</code> type. <code>null</code> if this
    *           method does not return an object instance.
    */
   public void addDataMember(String  name,
                             int     type,
                             int     access,
                             boolean isStatic,
                             String  qname)
   {
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         cls.addVariable(name, type, access, isStatic, qname);
      }
   }
   
   /**
    * Searches the named class' hierarchy for the named data member. This
    * will match both static and instance members.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    *
    * @return   The token type of the member or -1 if no match was found.
    */             
   public int lookupDataMember(String cname, String name)
   {
      return lookupDataMember(cname, name, false);
   }
   
   /**
    * Searches the named class' hierarchy for the named data member.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   The token type of the member or -1 if no match was found.
    */             
   public int lookupDataMember(String cname, String name, boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() 
                                            : lookupClass(cname);
                                            
      int access = calcAccessMode(cname);
      
      return (cls == null) ? -1 : cls.lookupVariable(name, access, isStatic);
   }   
   
   /**
    * Searches the named class' hierarchy for the named data member and
    * returns the fully qualified class name if that member represents an
    * object instance.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a member within the local class definition.
    * @param    name
    *           Member name to lookup.
    * @param    isStatic    
    *           <code>true</code> if this is a static reference.
    *
    * @return   Fully qualified class name of the the object instance
    *           represented by the named variable, or <code>null</code> if no
    *           such variable exists. 
    */             
   public String lookupDataMemberClass(String  cname,
                                       String  name,
                                       boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() 
                                            : lookupClass(cname);

      int access = calcAccessMode(cname);
      
      return (cls == null) ? null : cls.lookupVariableClassName(name,
                                                                access,
                                                                isStatic);
   }   
   
   /**                                                          
    * Searches the named class' hierarchy for the named method. This will
    * match any method (static or instance) of the given name.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a method call within the local class definition.
    * @param    method
    *           Method name to lookup.
    *
    * @return   The return type of the method or -1 if no match was found.
    */             
   public int lookupObjectMethod(String cname, String method)
   {
      return lookupObjectMethod(cname, method, false);
   }   
   
   /**                                                          
    * Searches the named class' hierarchy for the named method.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a method call within the local class definition.
    * @param    method
    *           Method name to lookup.
    * @param    isStatic
    *           If <code>true</code>, only static methods will be looked up.
    *           Otherwise any method can be returned.
    *
    * @return   The return type of the method or -1 if no match was found.
    */             
   public int lookupObjectMethod(String cname, String method, boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() 
                                            : lookupClass(cname);

      int access = calcAccessMode(cname);
      
      return (cls == null) ? -1 : cls.lookupMethod(method, access, isStatic);
   }   
   
   /**
    * Searches the named class' hierarchy for the named method and returns
    * the fully qualified class name if that method returns an object
    * instance.
    *
    * @param    cname
    *           Class name to search within or <code>null</code> if this is
    *           a method call within the local class definition.
    * @param    method
    *           Method name to lookup.
    *
    * @return   Fully qualified class name of the the object instance
    *           represented by the named method, or <code>null</code> if no
    *           such method exists. 
    * @param    isStatic    
    *           <code>true</code> if this is a static reference.
    */
   public String lookupObjectMethodClass(String  cname,
                                         String  method,
                                         boolean isStatic)
   {
      ClassDefinition cls = (cname == null) ? getCurrentClassDef() 
                                            : lookupClass(cname);

      int access = calcAccessMode(cname);
      
      return (cls == null) ? null : cls.lookupMethodClassName(method,
                                                              access,
                                                              isStatic);
   }   
   
   /**
    * Adds an attribute or method keyword token type and the associated token 
    * type of the data type (for attributes) or return type (for methods) to
    * the dictionary of attributes/methods.
    * <p>
    * This is a flat dictionary (there is no scoping).  The keyword's token
    * type is used as the key and the result token type is the value.
    * <p>
    * This form of the method is NOT suitable for adding attributes or
    * methods with <code>ATTR_CLASS</code> or <code>METH_CLASS</code> result
    * types. This is due to the fact that there is no mechanism to
    * associate a class name with the resulting attribute or method.
    *
    * @param    ktype
    *           Keyword token type to be used as the key in the dictionary.
    * @param    rtype
    *           The token type representing the attribute data type or the 
    *           method return type as defined by the Lexer and Parser.
    */   
   public void addAttributeOrMethod(int ktype, int rtype) 
   {
      addAttributeOrMethod(ktype, rtype, null);
   }

   /**
    * Adds an attribute or method keyword token type and the associated token 
    * type of the data type (for attributes) or return type (for methods) to
    * the dictionary of attributes/methods.
    * <p>
    * This is a flat dictionary (there is no scoping).  The keyword's token
    * type is used as the key and the result token type is the value.
    *
    * @param    ktype
    *           Keyword token type to be used as the key in the dictionary.
    * @param    rtype
    *           The token type representing the attribute data type or the 
    *           method return type as defined by the Lexer and Parser.
    * @param    cls
    *           Fully qualified name of the class (user defined type) that
    *           is returned from this method. Must be <code>null</code> if
    *           the resultType is not <code>ATTR_CLASS</code> or
    *           <code>METH_CLASS</code>.
    */   
   public void addAttributeOrMethod(int ktype, int rtype, String cls) 
   {
      attributes.put(ktype, rtype);
      
      if (cls != null)
         attrTypes.put(ktype, cls);
   }

   /**
    * Searches the attribute and method dictionary for a token type matching
    * the passed keyword.  The token type of the method return type or the
    * attribute data type is returned, if a match is found.  A token type
    * of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param    ktype
    *           Token type of the keyword to lookup.
    *
    * @return   The attribute or method token type that was found or -1 if no
    *           match was found.
    */             
   public int lookupAttributeOrMethod(int ktype)
   {
      Integer result = attributes.get(ktype);
      
      return result == null ? -1 : result.intValue();
   }   
   
   /**
    * Searches the attribute and method dictionary for the fully qualified
    * class name that is the actual type of an <code>ATTR_CLASS</code> or 
    * <code>METH_CLASS</code> matching the given keyword.
    *
    * @param    ktype
    *           Token type of the keyword to lookup.
    *
    * @return   The fully qualified class name that is the actual type 
    *           of an <code>ATTR_CLASS</code> or <code>METH_CLASS</code>.
    *           <code>null</code> if the given keyword is not a valid class
    *           attribute or method.
    */             
   public String lookupAttributeOrMethodClass(int ktype)
   {
      int ttype = lookupAttributeOrMethod(ktype);
      
      String result = null;
      
      if (ttype == ATTR_CLASS || ttype == METH_CLASS)
      {
         result = attrTypes.get(ktype);
      }
      
      return result;
   }   
   
   /**
    * Adds a label and a token type to the dictionary of labels.  This is
    * a special form of a scoped dictionary where there is only one label name
    * per scope (Progress does not allow lable aliases). The label name is
    * used as the key and the token type is the value.
    *
    * @param   name
    *          Label name.
    * @param   tokenType
    *          The label's token type as defined by the Lexer and Parser.
    */   
   public void addLabel(String name, int tokenType) 
   {
      labelDict.addScope(null);
      labelDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the label dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The label's token type that was found or -1 if no match was
    *          found.
    */             
   public int lookupLabel(String name)
   {
      return lookupWorkerExact(labelDict, name);
   }   
      
   /**
    * Deletes the current label from the dictionary of labels.  This is
    * a special form of a scoped dictionary where there is only one label name
    * per scope (Progress does not allow lable aliases). So, deleting the
    * current label is deleting the current scope.
    */   
   public void deleteLabel() 
   {
      labelDict.deleteScope();
   }

   /**
    * Adds a widget and a token type to the dictionary of widgets.
    * This dictionary supports scoping that matches the variable namespace.
    * The widget name is used as the key and the token type is the value.
    * <p>
    * Note that this is artificially separated from the variable namespace
    * for implementation simplicity, but in Progress there is only one
    * namespace that handles both variables and widgets.
    *
    * @param   name
    *          Widget name.
    * @param   tokenType
    *          The procedure's token type as defined by the Lexer and Parser.
    */   
   public void addWidget(String name, int tokenType) 
   {
      Variable var = new Variable(name, tokenType, null);
      
      addWrappedWorker(widgetDict,
                       false,
                       name,
                       (TokenDataWrapper) var);
   }

   /**
    * Searches the widget dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    * <p>
    * Note that this is artificially separated from the variable namespace
    * for implementation simplicity, but in Progress there is only one
    * namespace that handles both variables and widgets.
    *
    * @param   name
    *          Text string to match.
    * @return  The widget's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupWidget(String name)
   {
      return lookupWrappedType(lookupWrapped(widgetDict, name));
   }   
   
   /**
    * Adds a frame and a token type to the dictionary of frames. The frame
    * name is used as the key and the token type is the value.
    *
    * @param   name
    *          Frame name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addFrame(String name, int tokenType) 
   {
      frameDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the frame dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The frame's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupFrame(String name)
   {
      return lookupWorkerExact(frameDict, name);
   }   
   
   /**
    * Adds a menu or sub-menu and a token type to the dictionary of menus.
    * The menu or sub-menu name is used as the key and the token type is the
    * value.
    *
    * @param   name
    *          Menu or sub-menu name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addMenu(String name, int tokenType) 
   {
      menuDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the menu dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    * <p>
    * This supports both menus and sub-menus.
    *
    * @param   name
    *          Text string to match.
    * @return  The menu's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupMenu(String name)
   {
      return lookupWorkerExact(menuDict, name);
   }   
   
   /**
    * Adds a menu-item and a token type to the dictionary of menu-items.
    * The menu-item name is used as the key and the token type is the value.
    *
    * @param   name
    *          Menu-item name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addMenuItem(String name, int tokenType) 
   {
      menuItemDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the menu-item dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The menu-item's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupMenuItem(String name)
   {
      return lookupWorkerExact(menuItemDict, name);
   }   
   
   /**
    * Adds a stream and a token type to the dictionary of streams.
    * The stream name is used as the key and the token type is the value.
    *
    * @param   name
    *          Stream name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addStream(String name, int tokenType) 
   {
      streamDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the stream dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The stream's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupStream(String name)
   {
      return lookupWorkerExact(streamDict, name);
   }   
   
   /**
    * Adds a dataset and a token type to the dictionary of datasets.
    * The dataset name is used as the key and the token type is the value.
    *
    * @param   name
    *          Dataset name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addDataSet(String name, int tokenType) 
   {
      dsetDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the dataset dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The dataset's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupDataSet(String name)
   {
      return lookupWorkerExact(dsetDict, name);
   }   
   
   /**
    * Adds a data-source and a token type to the dictionary of data-sources.
    * The data-source name is used as the key and the token type is the value.
    *
    * @param   name
    *          Data-source name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addDataSource(String name, int tokenType) 
   {
      dsrcDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the data-source dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The data-source's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupDataSource(String name)
   {
      return lookupWorkerExact(dsrcDict, name);
   }   
   
   /**
    * Adds a data-relation and a token type to the dictionary of
    * data-relations. The data-relation name is used as the key and the token
    * type is the value.
    *
    * @param   name
    *          Data-relation name.
    * @param   tokenType
    *          The token type as defined by the Lexer and Parser.
    */   
   public void addDataRelation(String name, int tokenType) 
   {
      drelDict.addSymbol(false, name, new Integer(tokenType));
   }

   /**
    * Searches the data-relation dictionary for a symbol matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match.
    *
    * @param   name
    *          Text string to match.
    * @return  The data-relation's token type that was found or -1 if no match
    *          was found.
    */             
   public int lookupDataRelation(String name)
   {
      return lookupWorkerExact(drelDict, name);
   }   
   
   /**
    * Loads a list of databases into the schema dictionary.  This is the
    * way that non-default databases are 'connected' or added into the 
    * known namespace.
    *
    * @param    dblist
    *           List of database names to load.
    */
   public void loadSchemaDatabases(String[] dblist)
   {
      if (dblist == null || schemaDict == null)
         return;
     
      try
      {
         for ( int i = 0; i < dblist.length; i++ )
         {
            schemaDict.loadSchema(dblist[i]);
         }
      }
      catch (SchemaException excpt)
      {
         System.err.println(excpt.toString());
      }
   }
   
   /**
    * Adds a list of aliases (each to an associated database name) to the 
    * schema dictionary.
    *
    * @param    list
    *           A list of string arrays, each representing an alias and
    *           associated database to add to the schema dictionary.
    */
   public void loadAliases(List list)
   {
      if (schemaDict != null)
      {
         int len = list.size();
         
         for (int i = 0; i < len; i++)
         {
            String[] pair   = (String[]) list.get(i);
            boolean  result = schemaDict.createAlias(pair[0], pair[1]);
            
            if (!result)
            {
               // todo: yes, more logging here too
               System.err.println("createAlias() failed on " + pair[0] +
                                  " and database " + pair[1]);
            }
         }
      }
   }
   
   /**
    * Returns the shortest unique non-abbreviated buffer name for a field
    * or table reference.  This allows one to calculate the identical buffer
    * name from each reference, such that all references that use the same
    * buffer will share the same buffer name, even if the references are
    * different (due to abbreviations, field vs. table, qualified vs
    * unqualified) in the original source code.
    *
    * @param   name
    *          Text string to match, exactly as found in the original source
    *          code (abbreviated, qualified or inqualified). 
    * @param   field
    *          <code>true</code> if this is a field reference, 
    *          <code>false</code> if this is a table reference.
    *
    * @return  The shortest unabbreviated unique buffer name or
    *          <code>null</code> if no match was found.
    */
   public String lookupBufferName(String name, boolean field)
   {
      String result = null;
      
      try
      {
         if (schemaDict != null)
         {
            if (field)
            {
               result = schemaDict.getBufferForField(name);
            }
            else
            {
               result = schemaDict.getBufferForTable(name);
            }
         }
      }
      
      catch (AmbiguousSchemaNameException exc)
      {
         exc.printStackTrace();
      }
      
      return result;
   }
   
   /**
    * Returns the logical database name being used for a given table
    * reference.  The resulting name may be different from the database
    * portion of the fully qualified schema name as returned by the
    * {@link #lookupFullTableName} since the same schema can be referenced
    * as multiple database instances.
    *
    * @param   name
    *          Text string to match, exactly as found in the original source
    *          code (abbreviated, qualified or inqualified). 
    *
    * @return  The logical database name or <code>null</code> if no match was
    *          found.
    */
   public String lookupDatabaseName(String name)
   {
      String result = null;
      
      try
      {
         if (schemaDict != null)
         {
            result = schemaDict.getLogicalDatabaseForTable(name);
         }
      }
      
      catch (AmbiguousSchemaNameException exc)
      {
         exc.printStackTrace();
      }
      
      return result;
   }
   
   /**
    * Returns the fully qualified name for a field matching the passed 
    * key.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of field name portion. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table name is the smallest possible 
    * abbreviation.  As such, no ambiguity is allowed in the name passed as 
    * an argument.  There must be only one match that is ever found in the
    * schema namespace.
    *
    * @param   name
    *          Text string to match, in either of the 3 possible forms.
    *
    * @return  The field's fully qualified schema name or <code>null</code>
    *          if no match was found.
    */
   public String lookupFullFieldName(String name)
   {
      String fqn = null;
      
      try
      {
         if (schemaDict != null)
         {
            fqn = schemaDict.getFullFieldName(name);
         }
      }
      catch (Exception exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return fqn;
   }   
   
   /**
    * Searches the schema dictionary for a field matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous field name or due to some other problem with the schema
    * dictionary.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    *
    * @param    name
    *           Text string to match, in any of the supported 3 forms.
    *
    * @return   The label's token type that was found or -1 if no match was
    *           found.
    */
   public int lookupField(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            type = schemaDict.getFieldDataType(name);
         }        
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }
   
   /**
    * Searches the schema dictionary for a field matching the passed 
    * key and returns a boolean reporting if it is a valid field or not.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    * <p>
    * <b>This method is safe to use inside a semantic predicate since it
    * will not throw an exception if the name is not found or if there is
    * ambiguity detected!</b>
    *
    * @param   name
    *          Text string to match, in any of the supported 3 forms.
    *
    * @return  true if the name is a valid field, false otherwise.
    */
   public boolean isField(String name)
   {
      return schemaDict == null ? false : schemaDict.isField(name);
   } 
   
   /**
    * Returns the fully qualified name for a table matching the passed 
    * key.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms of input are accepted.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames portion. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table name is the smallest possible 
    * abbreviation.  As such, no ambiguity is allowed in the name passed as 
    * an argument.  There must be only one match that is ever found in the
    * schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    *
    * @param   name
    *          Text string to match, in either of the 2 possible forms.
    *
    * @return  The table's fully qualified schema name or <code>null</code>
    *          if no match was found.
    */
   public String lookupFullTableName(String name)
   {
      String fqn = null;
      
      try
      {
         if (schemaDict != null)
         {
            fqn = schemaDict.getFullTableName(name);
         }
      }
      catch (Exception exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return fqn;
   }   
   
   /**
    * Searches the schema dictionary for a table matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous table name or due to some other problem with the schema
    * dictionary.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms of input are accepted.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames portion. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table name is the smallest possible 
    * abbreviation.  As such, no ambiguity is allowed in the name passed as 
    * an argument.  There must be only one match that is ever found in the
    * schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    *
    * @param    name
    *           Text string to match, in either of the 2 possible forms.
    *
    * @return   The table's token type that was found or -1 if no match was
    *           found.
    */
   public int lookupTable(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            type = schemaDict.getTableType(name);
         }
      }
      catch (IllegalArgumentException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }
   
   /**
    * Searches the schema dictionary for a table matching the passed 
    * key and returns a boolean reporting if it is a valid table or not.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    * <p>
    * <b>This method is safe to use inside a semantic predicate since it
    * will not throw an exception if the name is not found or if there is
    * ambiguity detected!</b>
    *
    * @param   name
    *          Text string to match, in either of the supported 2 forms.
    *
    * @return  true if the name is a valid field, false otherwise.
    */
   public boolean isTable(String name)
   {
      return schemaDict == null ? false : schemaDict.isTable(name);
   } 
   
   /**
    * Searches the schema dictionary for a field matching the passed 
    * name and then will return the token type of the backing construct
    * (table, temp-table or work-table) to which this record belongs.  Note
    * that a token type of -1 is invalid in the Lexer and Parser, so this
    * value is used to denote a search that found no match, or one that
    * failed due to an ambiguous table name or due to some other problem with
    * the schema dictionary.
    * <p>
    * Schema field names can come in 3 forms:
    * <ul>
    *    <li> unqualified (e.g. address1)
    *    <li> partially qualified with a tablename (e.g. customer.address1)
    *    <li> fully qualified with database and table (e.g. 
    *         salesapp.customer.address1)
    * </ul>
    * <p>
    * All 3 forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    *
    * @param   name
    *          Text string to match, in any of the supported 3 forms.
    *
    * @return  The backing record's token type that was found or -1 if no
    *          match was found.
    */
   public int lookupFieldRecordType(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            type = schemaDict.getFieldRecordType(name);
         }
      }
      catch (IllegalArgumentException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }    
   
   /**
    * Searches the schema dictionary for a record matching the passed 
    * name and then will return the token type of the backing construct
    * (table, temp-table or work-table) to which this record belongs. By
    * definition, if one passes a name that references a table, temp-table
    * or work-table directly, then this method returns the same token type
    * value as can be obtained via {@link #lookupTable}. However, if the
    * name references a buffer, then this method will return the type of
    * the backing construct to which the buffer refers.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous table name or due to some other problem with schema
    * dictionary.
    * <p>
    * Schema table names can come in 2 forms:
    * <ul>
    *    <li> unqualified, table name only (e.g. customer)
    *    <li> fully qualified with database and table (e.g. salesapp.customer)
    * </ul>
    * <p>
    * Both forms are accepted as input.  All searches are done on 
    * a case-insensitive basis.  In addition, the Progress 4GL schema
    * supports optional abbreviation of tablenames and fieldnames. These 
    * abbreviations are valid as long as they represent a unique match. The
    * smallest unique match to a table or field name is the smallest 
    * possible abbreviation.  As such, no ambiguity is allowed in the 
    * name passed as an argument.  There must be only one match that is
    * ever found in the schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create buffers, temp-tables and work-tables that all
    * are interchangeable with table names.  This represents a dynamic
    * table namespace which must appear as a single unified space.
    *
    * @param   name
    *          Text string to match, in either of the supported 2 forms.
    *
    * @return  The backing record's token type that was found or -1 if no
    *          match was found.
    */
   public int lookupTableRecordType(String name)
   {
      int type = -1;
      
      try
      {
         if (schemaDict != null)
         {
            type = schemaDict.getTableRecordType(name);
         }
      }
      catch (IllegalArgumentException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
      
      return type;
   }    
   
   /**
    * Searches the schema dictionary for a database matching the passed 
    * key.  The token type of the match is returned.  Note that a token
    * type of -1 is invalid in the Lexer and Parser, so this value is used
    * to denote a search that found no match, or one that failed due to an
    * ambiguous database name or due to some other problem with the schema
    * dictionary.
    * <p>
    * Schema database names are always unqualified (by '.' characters) and
    * cannot be abbreviated.  All searches are done on a case-insensitive
    * basis.  No ambiguity is allowed in the name passed as an argument.
    * There must be only one match that is ever found in the schema namespace.
    * <p>
    * In addition to the predefined table names from the schema, the user
    * may define or create database name aliases that are interchangeable
    * with database names.  This allows a procedure written to use a specific
    * database name to use (at runtime) a database of a different name that 
    * has an identical or at least compatible schema.  This represents a
    * dynamic namespace which must appear as a single unified space.
    * <p>
    * <b>At this time, this method does not support the lookup of aliases
    * since the dictionary does not provide such support.</b>
    * <p> 
    * <b>This method is safe to use inside a semantic predicate since it
    * will not throw an exception if the name is not found or if there is
    * ambiguity detected!</b>
    *
    * @param    name
    *           Text string to match.
    *
    * @return   The database's token type that was found or -1 if no match
    *           was found.
    */
   public int lookupDatabase(String name)
   {
      int type = -1;
      
      if (schemaDict != null && schemaDict.isDatabase(name))
         type = DATABASE;
      
      return type;
   }   
      
   /**
    * Allows the addition of a new table of any type (table, buffer,
    * temp-table or work-table) to the schema namespace.  Additions are
    * made either to the global namespace (which is the same as the external
    * procedure scope) or to the current internal procedure or function
    * scope.  This is user controlled by a parameter to this method call. 
    * <p>
    * The database to which these are added is <code>null</code> and is
    * picked up by context in the schema dictionary or the table is
    * simply not associated with a specific database (true for temp-tables
    * and work-tables at least).
    * <p>
    * Any of these token types are accepted as the 2nd parameter:
    * <p>
    * <ul>
    *    <li> TABLE (this should never really occur)
    *    <li> BUFFER
    *    <li> TEMP_TABLE
    *    <li> WORK_TABLE
    * </ul>
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table (for the further addition of fields), is returned as a
    * type <code>Object</code>.  Field additions will require this instance
    * reference to be passed as a parameter.  See {@link #addField}.
    *
    * @param    tablename
    *           The name of the table to add.
    * @param    type
    *           The integer token type representing the type of table. 
    * @param    global
    *           Specifies if the table should be added to the global scope
    *           or to the current procedure/function scope.
    * @param    node
    *           The node that triggered the table creation.
    *
    * @return   The reference to the added table which is used in subsequent
    *           field adds.  Returns <code>null</code> on error.
    */
   public Object addTable(String  tablename,
                          int     type,
                          boolean global,
                          Aast    node)
   {
      if (tablename == null        ||
          type < BEGIN_RECORDTYPES ||
          type > END_RECORDTYPES)
         return null;
      
      return (Object) schemaDict.addTableEntry(null,
                                               tablename,
                                               type,
                                               global,
                                               node);
   }
      
   /**
    * Allows the addition of a new field of any Progress data type to a 
    * specific table in the schema namespace.  Additions are
    * always made to the global namespace, not to a specific scope.
    * <p>
    * Any of the valid field token types are accepted as the 3rd parameter
    * (e.g. <code>ProgressParserTokenTypes.FIELD_INT</code>).
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add the field.
    * @param    field
    *           The name of the field to add. 
    * @param    ftype
    *           The parser token type representing the field's data type.
    * @param    node
    *           Node in source file from which this new field creation was
    *           triggered.
    *
    * @return   The <code>Object</code> representing the added field or 
    *           <code>null</code> on failure.
    */
   public Object addField(Object table, String field, int ftype, Aast node)
   {
      if (table == null || field == null ||
          ftype < BEGIN_FIELDTYPES || ftype > END_FIELDTYPES)
         return null;
      
      return (Object) schemaDict.addFieldEntry((NameNode)table,
                                               node,
                                               field,
                                               ftype,
                                               true);
   }
      
   /**
    * Allows the addition of a new field of any Progress data type to a 
    * specific table in the schema namespace based on reference to an already
    * existing variable or field.  Additions are always made to the global
    * namespace, not to a specific scope.
    * <p>
    * Depending on the token type stored in the 3rd parameter, 2 possible
    * outcomes can occur:
    * <p>
    * <ol>
    *    <li> If the AST represents a variable type, then the field's options
    *         will be set to any non-default values stored in the variable 
    *         definition.
    *    <li> If the AST is a field, then the schema dictionary will create
    *         the field with its options set the same as the referenced
    *         field.
    * </ol>
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add the field.
    * @param    field
    *           The name of the field to add. 
    * @param    ast
    *           The AST node being used as a reference variable or field.
    *
    * @return   The <code>Object</code> representing the added field or 
    *           <code>null</code> on failure.
    */
   public Object addFieldLike(Object table, String field, Aast ast)
   {
      if (table == null || field == null || ast == null)
         return null;
      
      int     ftype   = ast.getType();
      int     extent  = getExtentSubscript(ast);
      String  oldname = ast.getText();
      boolean isFld   = ftype > BEGIN_FIELDTYPES && ftype < END_FIELDTYPES;
      
      if (isFld) 
      {
         NameNode node = null;
         
         // this field should be LIKE another field, call the schema
         // dictionary to add the field.
         try
         {
            node = schemaDict.addFieldEntry((NameNode) table,
                                            ast,
                                            field,
                                            oldname,
                                            extent);
         }
         catch (SchemaException exc)
         {
            // TODO: log this instead of using stderr...
            exc.printStackTrace();
         }
         
         return node;
      }
      
      // OK, we are add a field like a variable, so we add the field and then
      // copy our variable options into it
      int type = -1;
      
      // the VAR_ type must be converted to the proper FIELD_ type
      switch (ftype)
      {
         case VAR_CHAR:
            type = FIELD_CHAR;
            break;
         case VAR_CLASS:
            type = FIELD_CLASS;
            break;
         case VAR_COM_HANDLE:
            type = FIELD_COM_HANDLE;
            break;
         case VAR_DATE:
            type = FIELD_DATE;
            break;
         case VAR_DATETIME:
            type = FIELD_DATETIME;
            break;
         case VAR_DATETIME_TZ:
            type = FIELD_DATETIME_TZ;
            break;
         case VAR_DEC:
            type = FIELD_DEC;
            break;
         case VAR_HANDLE:
            type = FIELD_HANDLE;
            break;
         case VAR_INT:
            type = FIELD_INT;
            break;
         case VAR_INT64:
            type = FIELD_INT64;
            break;
         case VAR_LOGICAL:
            type = FIELD_LOGICAL;
            break;
         case VAR_RAW:
            type = FIELD_RAW;
            break;                     
         case VAR_RECID:
            type = FIELD_RECID;
            break;
         case VAR_ROWID:
            type = FIELD_ROWID;
            break;
         default:
            // this MUST be at the end of the list of variables, it will
            // match LONGCHAR and MEMPTR which cannot be created as a field
            if (ftype > BEGIN_VARTYPES && ftype < BEGIN_VARTYPES)
            {
               String err = String.format("Unexpected variable type %s!",
                                          ast.getDescriptiveTokenText());
               throw new RuntimeException(err);
            }
            break;
      }      
            
      if (type == -1)
         return null;
      
      Variable var = (Variable) lookupWrapped(varDict, oldname);
      NameNode node = null;
      try
      {
         schemaDict.addFieldEntry((NameNode) table,
                                  ast,
                                  field,
                                  type,
                                  var.getOptions(extent));
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
      
      return node;
   }
            
   /**
    * Inspect the field definition and write any options into the field 
    * object.
    *
    * @param    name
    *           The name of the field which must have its options set.
    * @param    ast
    *           The tree to inspect which defines a Progress field.
    */
   public void setFieldOptions(String name, Aast ast)
   {
      try
      {
         schemaDict.setFieldProperties(name, ast);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stderr...
         exc.printStackTrace();
      }      
   }      
      
   /**
    * Allows the addition of a list of fields (as defined in an existing
    * table) to a specific table in the schema namespace.  Additions are
    * always made to the global namespace, not to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    tablename
    *           The name of the table whose fields should be added. 
    */
   public void addFieldsFrom(Object table, String tablename)
   {
      try
      {
         schemaDict.addFieldEntries(tablename, (NameNode)table);
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
   }
   
   /**
    * Adds an index from the original source table into the temp-table that
    * is passed in.  Additions are always made to the global namespace, not
    * to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    node
    *           Node in source file from which this new field creation was
    *           triggered.
    * @param    name
    *           The name of the index to add. 
    * @param    primary
    *           Whether or not the <code>PRIMARY</code> keyword was present.
    */
   public void addIndexFrom(Object  table,
                            Aast    node,
                            String  name,
                            boolean primary)
   {
      if (table == null || name == null)
      {
         // nothing to do
         return;
      }
      
      // add "use index" processing to schema dictionary
      try
      {
         schemaDict.copyIndex((NameNode) table, node, name, primary);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
            
   /**
    * Adds an index as defined in the source code into the temp-table that
    * is passed in.  Additions are always made to the global namespace, not
    * to a specific scope.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference in which to add fields.
    * @param    name
    *           The name of the index to add. 
    * @param    ast
    *           The AST storing the nodes that define the index.
    */
   public void addIndex(Object table, String name, Aast ast)
   {
      if (table == null || name == null || ast == null)
      {
         // nothing to do
         return;
      }
      
      try
      {
         // add a new index for the given table by the given name and
         // set the options based on our AST
         schemaDict.addIndex((NameNode) table, name, ast);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
                        
   /**
    * Sets specific conditions on a temp-table definition based on
    * encountering certain tokens in the source code.  The following tokens
    * are known: <code>NO-UNDO, VALIDATE, RCODE-INFORMATION</code>.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference which is completed.
    * @param    token
    *           The type of the option that was encountered in the source
    *           code.
    * @param    node
    *           The node that triggered the table creation.
    */
   public void setTableOption(Object table, int token, Aast node)
   {
      Aast tableAst = schemaDict.getTable((NameNode) table);
      try
      {
         Aast prop = new ProgressAst();
         prop.setType(token);
         prop.setLine(node.getLine());
         prop.setColumn(node.getColumn());
         
         switch (token)
         {
            case KW_NO_UNDO:
            {
               // set NO-UNDO option for this table
               prop.setText("no-undo");
               break;
            }
            
            case KW_VALIDATE:
               // set VALIDATE option for this table
               prop.setText("validate");
               break;
               
            case KW_RCOD_INF:
            {
               // set RCODE-INFORMATION option for this table
               prop.setText("rcode-information");
               break;
            }
         }
         
         schemaDict.setTableProperty(tableAst, prop);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }                        
                        
   /**
    * Notifies the schema dictionary that the end of a temp-table definition
    * has been reached. This allows certain default processing to be executed
    * as needed.
    * <p>
    * In order to hide the schema dictionary interfaces from the caller,
    * the <code>NameNode</code> object that can be used to reference
    * the table is required.  This is an <code>Object</code> returned from a
    * previous call to {@link #addTable}.  This rule casts that to the
    * proper type before calling the dictionary.
    *
    * @param    table
    *           The table reference which is completed.
    */
   public void finalizeTableDefine(Object table)
   {
      try
      {
         // notify the schema dictionary of the "end of temp-table
         // definition"
         schemaDict.finalizeTableDefine((NameNode) table);
      }
      catch (SchemaException exc)
      {
         exc.printStackTrace();
      }
   }
   
   /**
    * Adds a record scope to the schema namespace to the top of the stack
    * of scopes.  Any tables that are promoted (see {@link #promoteTable})
    * within this scope are given precedence in making their field names
    * unambiguous when compared to fields of the same name in tables that
    * are in scopes that are lower on the stack.
    *
    * @param    mark
    *           <code>true</code> to mark the associated scope as one which
    *           is an internal procedure, user-defined function or trigger.
    *           <code>false</code> for any other case including external
    *           procedures and inner blocks (<code>DO, REPEAT or FOR</code>).
    */
   public void addSchemaScope(boolean mark)
   {
      if (schemaDict != null)
      {
         schemaDict.addScope(mark);
      }
   }
      
   /**
    * Deletes the current record scope in the schema namespace (the top of
    * the stack of scopes).  Any tables that were promoted (see
    * {@link #promoteTable}) into this scope are naturally popped off the
    * stack by this action if <code>propagate</code> is <code>false</code>.
    * Otherwise the namespace updates made in this scope will be copied to
    * the namespaces of the enclosing scope.
    *
    * @param    propagate
    *           <code>true</code> to force namespace propagation to the
    *           enclosing scope.
    */
   public void deleteSchemaScope(boolean propagate)
   {
      try
      {
         if (schemaDict != null)
         {
            schemaDict.removeScope(propagate);
         }
      }
      
      catch (SchemaException exp)
      {
         // ignore it, we don't care!
      }
   }      
      
   /**
    * Globally enable or disable schema table name promotion.
    * 
    * @param    allow
    *           <code>true</code> to enable, <code>false</code> to disable.
    */
   public void allowPromotion(boolean allow)      
   {
      this.allow = allow;
   }
   
   /**
    * Promotes a table (and all its fields) into the current top scope of
    * the schema dictionary's stack of scopes.  Any tables that are promoted
    * within this scope are given precedence in making their field names
    * unambiguous when compared to fields of the same name in tables that
    * are in scopes that are lower on the stack.
    * <p>
    * Scopes are created using {@link #addSchemaScope} and are deleted using 
    * {@link #deleteSchemaScope}.
    * 
    * @param    table
    *           The name of the table (optionally qualified) to promote.
    * @param    force
    *           Override the default promotion behavior and force a refresh
    *           of the promoted table even if it was already promoted
    *           previously.
    * @param    noProp                   
    *           <code>true</code> to force any promoted nodes to have their
    *           no propagate flag set. This disables propagation of these
    *           nodes across the current scope boundary.
    */
   public void promoteTableName(String table, boolean force, boolean noProp)      
   {
      try
      {
         if (allow)
         {
            schemaDict.promoteTable(table, force, noProp);
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
   }
      
   /**
    * Promotes a table (and all its fields) into the current top scope of
    * the schema dictionary's stack of scopes.  Any tables that are promoted
    * within this scope are given precedence in making their field names
    * unambiguous when compared to fields of the same name in tables that
    * are in scopes that are lower on the stack.
    * <p>
    * Scopes are created using {@link #addSchemaScope} and are deleted using 
    * {@link #deleteSchemaScope}.
    * 
    * @param    table
    *           The <code>Object</code> reference representing a table
    *           that had been previously added to the schema dictionary.
    * @param    force
    *           Override the default promotion behavior and force a refresh
    *           of the promoted table even if it was already promoted
    *           previously.
    * @param    noProp                   
    *           <code>true</code> to force any promoted nodes to have their
    *           no propagate flag set. This disables propagation of these
    *           nodes across the current scope boundary.
    */
   public void promoteTable(Object table, boolean force, boolean noProp)      
   {
      NameNode tbl = (NameNode) table;
      try
      {
         if (allow)
         {
            schemaDict.promoteTable(tbl.getName(), force, noProp);
         }
      }
      catch (SchemaException exc)
      {
         // TODO: log this instead of using stdout...
         exc.printStackTrace();
      }
   }
   
   /**
    * Add a handle name to the list of possible active-x targets.
    *
    * @param    target
    *           The handle name.
    */
   public void setPossibleActiveX(String target)
   {
      possibleActiveX.add(target.toLowerCase());
   }
      
   /**
    * Check if a handle name is in the list of possible active-x targets. The
    * name lookup is done on a case-insensitive basis.
    *
    * @param    possible
    *           The handle name to check.
    *
    * @return   <code>true</code> if the name is in the list of possible
    *           active-x targets.
    */
   public boolean isPossibleActiveX(String possible)
   {
      return possibleActiveX.contains(possible.toLowerCase());
   }
      
   /**         
    * Prints a report to <code>stderr</code> with the listing of each
    * dictionary.  One record per dictionary entry will be printed and
    * there will be a section for each namespace type.
    */
   public void dump()
   {
      System.err.println("\n-------------Keyword Listing-------------");      
      kwDict.dump();
      
      System.err.println("\n-------------Variable Listing-------------");
      varDict.dump();
      
      System.err.println("\n-------------Function Listing-------------");
      funcDict.dump();
      
      System.err.println("\n-------------Procedure Listing-------------");
      procDict.dump();
      
      System.err.println("\n---------------Label Listing---------------");
      labelDict.dump();
      
      System.err.println("\n--------------Widget Listing--------------");
      widgetDict.dump();
      
      System.err.println("\n---------------Frame Listing---------------");
      frameDict.dump();
      
      System.err.println("\n---------------Menu Listing---------------");
      menuDict.dump();
      
      System.err.println("\n-------------Menu Item Listing-------------");
      menuItemDict.dump();
      
      System.err.println("\n--------------Stream Listing--------------");
      streamDict.dump();
   }
   
   /**         
    * Prints a report to <code>schema.dmp</code> with the listing of all
    * current contents of the Schema Dictionary.
    */
   public void dumpSchema()
   throws IOException,
          FileNotFoundException
   {
      FileOutputStream fout = null;
      PrintStream      ps   = null;
      try
      {
         if (schemaDict != null)
         {
            File file = new File("schema.dmp");
            fout = new FileOutputStream(file);
            ps = new PrintStream(fout);
            schemaDict.dump(ps);
            System.out.println(
               file.getAbsolutePath() + " dump file created");
         }
      }
      finally
      {
         if (ps != null)
         {
            ps.close();
         }
         else if (fout != null)
         {
            fout.close();
         }
      }
   }
      
   /**
    * Returns the token type from the passed wrapped object or -1 for an
    * is invalid {@link TokenDataWrapper} object.
    *
    * @param   wrapped
    *          The object from which to obtain the token type.
    *
    * @return  The token type that was found or -1 for an invalid wrapper
    *          object.
    */    
   public int lookupWrappedType(TokenDataWrapper wrapped)
   {
      return wrapped != null ? wrapped.getTokenType() : -1;
   }
   
   /**
    * Initialize the maps of all possible class and interface definitions that
    * exist in the project. This includes all <code>.cls</code> files which can
    * be found via the <code>PROPATH</code> as well as all Progress built-ins
    * and all .NET classes/interfaces that can be accessed.
    * <p>
    * These files will be listed from the directories represented by the
    * {@link #propath}, {@link #OO4GL_PATH}, {@link #DOTNET_PATH} and the
    * {@link #ASSEMBLY_PATH}.
    * <p>
    * If file-system case-sensitivity is <code>false</code> then all map keys
    * will be lowercased.
    *
    * @param    propath
    *           The propath to scan for initialization. The propath can change
    *           with each instantiation of this class.  If it changes then
    *           the propath class mappings must be reinitialized. This may be
    *           <code>null</code> in which case the map will be empty.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   private static synchronized void initPossibleClasses(String[] propath,
                                                        boolean  caseSens)
   {
      if (propathCls == null)
      {
         propathCls = new HashMap<String, String>();
         XmlHelper.loadStringMap(propathCls, CLASS_MAP_FILE); 
         
         // this is the dynamic scanning approach, it is horrifically SLOW
         // which is why we don't use it; the code must also add this check
         // in the above if stmt: " || propath != lastPropath)"
         /*
         lastPropath = propath;
         
         if (propath != null)
         {
            boolean curDir = false;
            
            for (int i = 0; i < propath.length; i++)
            {
               if (propath[i] != null)
               {
                  String srch = propath[i];
                  
                  if (srch.length() == 0 || srch.equals("."))
                  {
                     // only search the current directory once
                     if (curDir)
                        continue;
                     
                     srch   = ".";
                     curDir = true;
                  }
                  
                  createClassMappings(propathCls,
                                      getClassFilenames(srch, caseSens),
                                      srch,
                                      caseSens);
               }
            }
         }
         */
      }
   
      // add the built-ins, .net and custom fakeout classes/interfaces; this
      // only happens the first time this method is called
      // TODO: this assumes the caseSens value is the same project-wide AND
      //       that the correct value is passed in the first time the
      //       class is instantiated (which can be by the SchemaLoader)
      if (oo4glCls == null)
      {
         oo4glCls = new HashMap<String, String>();
         createClassMappings(oo4glCls,
                             getClassFileList(OO4GL_PATH, caseSens),
                             OO4GL_PATH,
                             caseSens);
      }
      if (dotnetCls == null)
      {
         dotnetCls = new HashMap<String, String>();
         createClassMappings(dotnetCls,
                             getClassFileList(DOTNET_PATH, caseSens),
                             DOTNET_PATH,
                             caseSens);
      }
      if (assemblyCls == null)
      {
         assemblyCls = new HashMap<String, String>();
         createClassMappings(assemblyCls,
                             getClassFileList(ASSEMBLY_PATH, caseSens),
                             ASSEMBLY_PATH,
                             caseSens);
      }
   }
   
   /**
    * Store each element of the list as a key in the map and create a filename
    * as the mapped value (using the path + the key). The keys will be
    * lowercased if case-insensitive matching is configured.
    *
    * @param    map
    *           The map to modify.
    * @param    list
    *           The elements to add.
    * @param    path
    *           The root path in which all elements reside.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   private static void createClassMappings(Map<String, String> map,
                                           String[]            list,
                                           String              path,
                                           boolean             caseSens)
   {
      int begin = path.length();
      
      for (int i = 0; i < list.length; i++)
      {
         int end = list[i].length() - 4;
         String entry = list[i].substring(begin, end);
         map.put(caseSens ? entry : entry.toLowerCase(), list[i]);
      }
   }
   
   /**
    * List the 4GL class/interface files in a given path.
    *
    * @param    path
    *           The directory to search.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    *
    * @return   The list of files that exist.  If there are no files, the
    *           array will have a 0 length.
    */
   private static String[] getClassFileList(String path, boolean caseSens)
   {
      File dir = new File(path);
      
      if (!dir.exists() || !dir.isDirectory())
      {
         return new String[0];
      }
      
      FileList fl = new FileSpecList(dir,
                                     CLASS_SPLAT,
                                     true,
                                     caseSens);
      return fl.listFilenames();
   }
   
   /**
    * List the 4GL class/interface files in a given path.
    *
    * @param    path
    *           The directory to search.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    *
    * @return   The list of files that exist.  If there are no files, the
    *           array will have a 0 length.
    */
   private static ArrayList<File> getClassFiles(String path, boolean caseSens)
   {
      ArrayList<File> results = new ArrayList<File>();
      
      scan(new File(path), results, caseSens);
      
      return results;
   }
   
   /**
    * Core implementation of the listing algorithm which uses the
    * <code>File</code> class to make a complete list of the directory 
    * contents, then this list is used to build a filtered list of just
    * those files that match the user-defined regular expression. Note
    * that if recursion was previously requested (on construction) by the
    * caller, this method will recursively call itself for every directory
    * it encounters in the list.  Only files will be returned in the results
    * list and all results are added to the results list that is passed as
    * a parameter.
    *
    * @param    dir
    *           The directory in which to list all files.
    * @param    results
    *           The list to which all matching files must be appended.
    * @param    caseSens
    *           <code>true</code> if the legacy system had a case-sensitive
    *           file-system.
    */
   private static void scan(File dir, ArrayList<File> results, boolean caseSens)
   {
      // we don't use a FilenameFilter here because this would filter out
      // all directories that don't match the regular expression, but this
      // would not be correct, we need to see all directories and we will
      // manually filter the filenames down ourselves
      File[] rawResults  = dir.listFiles();
      
      // separate file and directory results
      for (int i = 0; i < rawResults.length; i++)
      {
         if (rawResults[i].isDirectory())
         {
            // recursively call ourselves to process this subdirectory
            scan(rawResults[i], results, caseSens);
         }
         else
         {
            // only accept matches that meet the user's criteria (regular
            // expression)
            if (rawResults[i].getName().endsWith(CLASS_EXT))
            {
               // add the file to the results list
               results.add(rawResults[i]);
            }
         }
      }
   }
   
   /**
    * Multi-level lookup routine for processing searches on variable and
    * function dictionaries based on
    * the {@link com.goldencode.p2j.util.ScopedSymbolDictionary} class. All 
    * searches resolve to a symbol that exactly matches the passed key or to
    * a built-in function or variable that is represented by a language
    * keyword.  The token type of the match is returned.  Note that a token  
    * type of -1 is invalid in the Lexer and Parser, so this value is used to 
    * denote a search that found no match.  Abbreviation processing is
    * handled for built-ins ONLY and is naturally done by usage of the 
    * keyword dictionary in the first phase of the lookup process.  The
    * second lookup phase uses any returned keyword to modify the search or 
    * if no keyword is found, then a standard exact search is executed.
    * <p> 
    * The algorithm checks for a keyword match.  If not, a simple actual text
    * lookup is done based on the exact text found in the source. If there is
    * a keyword match, then if the keyword is reserved, the keyword's full
    * text is used in the name lookup.  The result would match the preloaded 
    * built-in name.  If the keyword match is with an unreserved keyword,
    * then the exact text from the source is used for a first lookup.  If this
    * does not succeed, then a second lookup using the full keyword text is
    * used.  This approach allows unreserved keywords to be used as user-
    * defined names that hide built-in names if they exist.  However the
    * built-in names are still found if there is no user-defined name.
    * <p>
    * This usage of the full keyword text for lookup of built-ins is required
    * since only the full text definition exists in the variable dictionary, 
    * rather than the full set of abbreviations.  Instead the abbreviation
    * matching is handled by the keyword dictionary and this method is aware
    * of the difference.      
    *
    * @param   dict
    *          <code>ScopedSymbolDictionary</code> instance to search.
    * @param   actual
    *          Text string to match.
    * @return  The label's token type that was found or -1 if no match was
    *          found.
    * @see     #lookupVariable
    * @see     #lookupFunction
    */
   private int lookupWorker(ScopedSymbolDictionary dict, String actual)
   {
      Keyword kw    = lookupKeyword(actual);
      int     type  = -1;
      
      if (kw != null)
      {
         // the actual text matches a keyword
         if (kw.isReserved())
         {
            // since this is a reserved keyword, this must be a 
            // built-in variable/function
            type = lookupWorkerExact(dict, kw.getFullText());
         }
         else
         {
            // this is an unreserved keyword, this MAY be a 
            // built-in variable/function BUT we must allow a user-defined
            // name to take precedence (note that if the actual text is
            // is the same as the full keyword text, then we may still
            // get a match here that is the built-in, assuming no
            // user-defined name exists --> that is OK too)
            type = lookupWorkerExact(dict, actual);
            
            // check if there was a match to the actual text
            if (type == -1)
            {
               // match the full text of the keyword with the built-in
               type = lookupWorkerExact(dict, kw.getFullText());
            }
         }
      }
      else
      {
         // there is no keyword, so this must be a user-defined name
         type = lookupWorkerExact(dict, actual);
      }            
      
      return type;
   }
   
   /**
    * Primary worker routine for processing searches on dictionaries based on
    * the {@link com.goldencode.p2j.util.ScopedSymbolDictionary} class. All 
    * searches are for a symbol exactly matching the passed key.  The token
    * type of the match is returned.  Note that a token type of -1 is invalid 
    * in the Lexer and Parser, so this value is used to denote a search that
    * found no match.  No abbreviation processing will be provided!
    *
    * @param   dict
    *          <code>ScopedSymbolDictionary</code> instance to search.
    * @param   name
    *          Exact text string to match.
    * @return  The label's token type that was found or -1 if no match was
    *          found.
    */             
   private int lookupWorkerExact(ScopedSymbolDictionary dict, String name)   
   {
      int     type   = -1;
      Integer result = (Integer) dict.lookupSymbol(name);
      
      if (result != null)
      {
         type = result.intValue();
      }
      
      return type;
   }            
            
   /**   
    * Centralizes the process of setting an file-unique temporary index for
    * each new wrapped object and then adding that object to the passed-in
    * dictionary.
    *
    * @param    global
    *           Add to global scope if <code>true</code>.    
    * @param    name
    *           Variable name to add.
    * @param    wrapped
    *           Wrappered token object to add.
    */
   private void addWrappedWorker(ScopedSymbolDictionary ssd,
                                 boolean                global,
                                 String                 name,
                                 TokenDataWrapper       wrapped)
   {
      wrapped.setTempIndex(tempIdx);
      tempIdx++;
      ssd.addSymbol(global, name, wrapped);
   }
   
   /**
    * Given an AST which represents a field or variable which possibly has
    * an extent, check for a subscript and return it.  The assumed structure
    * of the AST, if it does have an extent subscript is:
    * <pre>
    *    FIELD/VAR
    *      |
    *      +--LBRACKET
    *           |
    *           +--EXPRESSION
    *                |
    *                +--NUM_LITERAL
    * </pre>
    * where the <code>FIELD/VAR</code> node is the <code>ast</code> parameter
    * passed to this method.
    * <p>
    * Currently, only the <code>NUM_LITERAL</code> token type is supported
    * for the subscript value.  Any other type of expression will raise an
    * exception.
    *
    * @param   ast
    *          Field or variable AST.
    *
    * @return  The 1-based subscript value from the <code>NUM_LITERAL</code>
    *          node depicted above, or 0 if there is no subscript.
    *
    * @throws  UnsupportedOperationException
    *          if a token type other than a simple <code>NUM_LITERAL</code>
    *          is provided for the subscript expression.
    * @throws  IllegalArgumentException
    *          if the <code>NUM_LITERAL</code> token cannot be parsed into an
    *          integral value.
    */
   private int getExtentSubscript(Aast ast)
   {
      int subscript = 0;
      
      if (ast.getFirstChild() != null)
      {
         Aast lbracket = ast.getImmediateChild(LBRACKET, null);
         if (lbracket != null)
         {
            Aast expr = lbracket.getChildAt(0);
            Aast numlit = expr.getImmediateChild(NUM_LITERAL, null);
            
            if (numlit == null)
            {
               String err = "Non-NUM_LITERAL extent value not supported " +
                            "in this context!";
               throw new UnsupportedOperationException(err);
            }
            
            try
            {
               subscript = Integer.parseInt(numlit.getText());
            }
            catch (NumberFormatException exc)
            {
               throw new IllegalArgumentException(
                  "Invalid extent subscript:  " + numlit.getText());
            }
         }
      }
      
      return subscript;
   }
   
   /**
    * Convert any "." characters (in a Progress qualified class or interface
    * name) to the current system's file name separator string.
    *
    * @param    name
    *           Qualified class or interface name.
    *
    * @return   Name that is safe to use with the current file system.
    */
   private String convertSeparator(String name)
   {
      return name.replaceAll("\\.", File.separator); 
   }
   
   /**
    * Search the PROPATH for a specific class or interface. The given name
    * should be a Progress style qualified name (package + class). This will
    * be converted to a file system name and searched for.
    *
    * @param    name
    *           Qualified class or interface name.
    * @param    type
    *           The type of the using clause that is being checked. This
    *           changes the search order and may drop some paths from the
    *           search when possible.
    * @param    result
    *           Object in which to store the search results (the file system
    *           name, the built-in flag and the .NET flag) when the search
    *           is successful. If nothing is found, this object is unchanged.
    *
    * @return   <code>true</code> if the class file was found.
    */
   private boolean findFile(String name, UsingType type, SearchResult result)
   {
      // we tried an approach based on FileSystemDaemon.search() but it is
      // slow, so we switched to a preloaded in-memory cache approach
      
      name = convertSeparator(name);
      
      Map<String, String>[] searchList = null;
      
      boolean[] builtin = new boolean[4];
      boolean[] dotnet  = new boolean[4];
      
      // choose the search paths based on the type
      switch (type)
      {
         case PROPATH:
            // no need to search .NET stuff
            searchList    = new HashMap[2];
            searchList[0] = propathCls;   builtin[0] = false; dotnet[0] = false;
            searchList[1] = oo4glCls;     builtin[1] = true;  dotnet[1] = false;
            break;
         case DOTNET:
            // the documentation says there is no need to search PROPATH but
            // working code has found Progress.* using stmts with the FROM
            // ASSEMBLY option, so we add the OO4GL stuff in as a last resort
            // TODO: should this be more general?
            searchList    = new HashMap[3];
            searchList[0] = dotnetCls;    builtin[0] = true;  dotnet[0] = true; 
            searchList[1] = assemblyCls;  builtin[1] = false; dotnet[1] = true; 
            searchList[2] = oo4glCls;     builtin[2] = true;  dotnet[2] = false;
            break;
         default:
            // search everything!
            searchList    = new HashMap[4];
            searchList[0] = propathCls;   builtin[0] = false; dotnet[0] = false;
            searchList[1] = oo4glCls;     builtin[1] = true;  dotnet[1] = false;
            searchList[2] = dotnetCls;    builtin[2] = true;  dotnet[2] = true;
            searchList[3] = assemblyCls;  builtin[3] = false; dotnet[3] = true;
            break;
      }
      
      String srch  = caseSens ? name : name.toLowerCase();
      
      for (int i = 0; i < searchList.length; i++)
      {
         Map<String, String> map = searchList[i];
         
         result.filename = map.get(srch);
         
         if (result.filename != null)
         {
            result.builtin = builtin[i];
            result.dotnet  = dotnet[i];
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Test if the given package exists in the PROPATH, or in the 4GL system
    * packages or in the .NET system namespaces.
    *
    * @param    pkg
    *           Package name to search for.
    *
    * @return   <code>true</code> if the package exists in the PROPATH.
    */
   private boolean pkgExists(String pkg)
   {
      pkg = convertSeparator(pkg);
      
      File file = null;
      
      for (int i = 0; i < propath.length; i++)
      {
         if (propath[i].length() == 0 || ".".equals(propath[i]))
         {
            // check a relative file in the current directory
            file = new File(pkg);
         }
         else
         {
            // check a relative file in the specified path
            file = new File(propath[i], pkg);
         }
         
         if (exists(file))
         {
            return true;
         }
      }
      
      // check if the package exists as part of the 4GL system
      file = new File(OO4GL_PATH, pkg);
      
      if (exists(file))
      {
         return true;
      }
      
      // check if the namespace for the system classes in .NET
      file = new File(DOTNET_PATH, pkg);
      
      if (exists(file))
      {
         return true;
      }
      
      // check if the namespace for the user assemblies in .NET
      file = new File(ASSEMBLY_PATH, pkg);
      
      if (exists(file))
      {
         return true;
      }
      
      return false;
   }
   
   /**
    * Implements a check for the file's existence while honoring the 
    * case-sensitive file-system flag.  If the original file-system was
    * case-sensitive, then a simple <code>exists()</code> method call is
    * made on the file object.  Otherwise a case-insensitive search is
    * undertaken.
    *
    * @param    file
    *           The file to check.
    *
    * @return   <code>true</code> if the file object exists.
    */
   private boolean exists(File file)
   {
      if (caseSens)
      {
         // does it exist?
         if (file.exists())
         {
            return true;
         }
      }
      else
      {
         return FileSystemDaemon.getCaseInsensitiveMatch(file, true) != null;
      }
      
      return false;
   }
   
   /**
    * Walk the given class or interface AST and load the contained class
    * or interface definition.
    *
    * @param    ast
    *           The tree to walk which contains the class or interface
    *           definition.
    */
   private void loadViaWalk(Aast ast)
   {
      Iterator iter    = ast.iterator();
      boolean  def     = false;
      boolean  builtin = ast.isAnnotation("builtin")
                            ? (Boolean) ast.getAnnotation("builtin")
                            : false;
      boolean  dotnet  = ast.isAnnotation("dotnet")
                            ? (Boolean) ast.getAnnotation("dotnet")
                            : false;
      
      while (iter.hasNext())
      {
         Aast next = (Aast) iter.next();
         
         switch (next.getType())
         {
            case KW_CLASS:
               if (next.getParent().getType() == CLASS_DEF)
               {
                  def = true;
                  
                  Aast symbol   = next.getChildAt(0);
                  Aast inherits = next.getImmediateChild(KW_INHERITS, null);
                  Aast imple    = next.getImmediateChild(KW_IMPLEMTS, null);
                  addClass(symbol, inherits, imple, builtin, dotnet);
               }
               break;
            case KW_INTERFAC:
               if (next.getParent().getType() == INTERFACE_DEF)
               {
                  def = true;
                  
                  Aast symbol = next.getChildAt(0);
                  addInterface(symbol, builtin, dotnet);
               }
               break;
            case CLASS_NAME:
               loadClass((String) next.getAnnotation("qualified"));
               break;
            case KW_METHOD:
               if (next.getParent().getType() == METHOD_DEF)
               {
                  String  name     = (String) next.getAnnotation("name");
                  long    mtype    = (Long) next.getAnnotation("return_type");
                  long    atype    = (Long) next.getAnnotation("access");
                  boolean isStatic = (Boolean) next.getAnnotation("static");
                  String  qname    = (String) next.getAnnotation("qualified");
                  
                  addObjectMethod(name,
                                  (int) mtype,
                                  (int) atype,
                                  isStatic,
                                  null,
                                  qname);
               }
               break;
            case DEFINE_VARIABLE:
            case DEFINE_PROPERTY:
            case DEFINE_EVENT:
               Aast   amode  = next.getImmediateChild(ACCESS_MODES, null);
               Aast   stic   = next.getImmediateChild(KW_STATIC, null);
               Aast   symbol = next.getImmediateChild(SYMBOL, null);
               Long   ttype  = (Long) next.getAnnotation("type");
               
               String qname  = null;
               
               if (next.isAnnotation("qualified"))
                  qname = (String) next.getAnnotation("qualified");
               
               addDataMember(symbol.getText(),
                             ttype.intValue(),
                             amode,
                             stic,
                             qname);          
               break;
         }
      }
      
      // the current file's definition is done, we don't have an ascent
      // rule here so just simulate the same result
      if (def)
         classDefinitionComplete();
   }
   
   /**
    * Calculates the access mode for method and data member searches of
    * class definitions, based on the presence and contents of the given
    * class name and the current context. By default the access mode will
    * be <code>KW_PUBLIC</code>. If this request is made within the context
    * of a class definition, then the access mode will be 
    * <code>KW_PRIVATE</code> if <code>cname</code> is <code>null</code> or
    * if <code>cname</code> is the same as the current class definition
    * name. The access mode will be <code>KW_PROTECTD</code> if the
    * <code>cname</code> is not <code>null</code> and it is the equivalent
    * of one of the parent class names in the current class definition's
    * inheritance chain.
    *
    * @param    cname
    *           Fully qualified class name being searched or <code>null</code>
    *           if the current class definition is to be searched.
    *
    * @return   The calculated access mode of <code>KW_PUBLIC</code>,
    *           <code>KW_PROTECTD</code> or <code>KW_PRIVATE</code>.
    */
   private int calcAccessMode(String cname)
   {
      int access = KW_PUBLIC;
      
      ClassDefinition cls = getCurrentClassDef();
      
      if (cls != null)
      {
         if (cname != null)
         {
            String current = cls.getName();
            
            if (cname.equalsIgnoreCase(current))
            {
               // inside a class def and explicitly referencing the current
               // class, this is a private search
               access = KW_PRIVATE;
            }
            else
            {
               ClassDefinition parent = cls.getParent();
               
               // search the parent hierarchy for a matching class name
               while (parent != null)
               {
                  String pname = parent.getName();
                  
                  if (cname.equalsIgnoreCase(pname))
                  {
                     // this is explicit access to a parent class from within
                     // a subclass definition, this is a protected search
                     access = KW_PROTECTD;
                     break;
                  }
                  
                  parent = parent.getParent();
               }
            }
         }
         else
         {
            // we are inside a class def and have no explicit class to
            // search, this is a private search of the current class'
            // hierarchy
            access = KW_PRIVATE;
         }
      }
      
      return access;
   }
   
   /** 
    * Describes the type of the USING specification.
    */
   private static enum UsingType
   {
      PROPATH, DOTNET, UNSPECIFIED
   };
   
   /**
    * Store the data in a USING statement.
    */
   private static class UsingSpec
   {
      /** Match specification (package/.NET namespace or full classname). */
      private String spec = null;
      
      /** Match type. */
      private UsingType type = UsingType.UNSPECIFIED;
      
      /**
       * Primary constructor.
       *
       * @param    spec
       *           Match specification (package/.NET namespace or full
       *           classname).
       * @param    type
       *           Match type.
       */
      public UsingSpec(String spec, UsingType type)
      {
         this.spec = spec;
         this.type = type;
      }
   }
   
   /**
    * Store the data resulting from a given file system search.
    */
   private static class SearchResult
   {
      /** File system object that was found (includes pathing). */
      private String filename = null;
      
      /** Fully qualified class or interface name relative to PROPATH. */
      private String clsname = null;
      
      /** Flag marking if this is a built-in Progress class. */
      private boolean builtin = false;
      
      /** Flag to differentiate between 4GL classes and .NET classes. */
      private boolean dotnet = false;
      
      /** Search hint. */
      private UsingType type = UsingType.UNSPECIFIED;
   }
}
